#include "AppHdr.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>

#ifdef _WIN32
#include <io.h>
#include <windows.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#include <pogo/pogo.h>
#include <pogo/console.h>
#include <pogo/readline.h>

#include "zlib.h"
#ifdef DS
#include <nds.h>
#include <nds/bios.h>
#endif

#include "defines.h"
#include "enum.h"

#include "tags.h"

extern "C" {
	void sdlkey_init();
	void key_init();
	void display_flush();
	int key_peek();
	int fatdev_init();
	void gbfsdev_init();
	void add_clickarea(int x, int y, int w, int h, int charw, uchar *);
	void activate_clickarea(int, int);
	int key_get_modifiers();
	void console_mark_square(int x, int y);

}

void render_panel();
int get_window();

int shift_pressed = 0;


static int dirpad = 0;
static int move = 0;
static int or_dirpad = 0;
//static int modif = 0;
static int a_key = -1;
static int key_rate = 0;
static int menu_mode = 0;
int menu_miny;
static int menu_maxy;
static int menu_x2;
static int my = -1;
static int mx = -1;
static int dir_mode = 0;
static int last_char = 0;

extern Font *font_main;

	//    1
	//   4 8
	//    2
	//                  012X456X89abcdef
char shortcuts[17][5] = {
	".,Zf",
	"krqz", // UP
	"j><p", // DOWN
	"    ",
	"hw'W", // LEFT
	"yEPt",
	"bync",
	"    ",
	"lsdx", // RIGHT
	"umM?",
	"n;eD",
	"    ",
	"    ",
	"<   ",
	">   ",
	"    "
};

static char *locked = NULL;

#define FL_DIRS 1
#define FL_THINGS 2
#define FL_LOCK 4

int sc_translate = 2;

int translate(char ch, int flags)
{
	int c = ch;

	if(flags & FL_LOCK)
	{
		switch(ch)
		{
			case 10:
				c = 243;
				break;
			case '\e':
				c = 244;
				break;
			//case ' ':
			//	c = 245;
			//	break;
		}
	}

	if((flags & FL_DIRS) && (sc_translate > 0))
	{
		switch(ch)
		{
			case 'k':
				c = 232; break;
			case 'j':
				c = 233; break;
			case 'l':
				c = 234; break;
			case 'h':
				c = 235; break;
			case 'y':
				c = 236; break;
			case 'u':
				c = 237; break;
			case 'n':
				c = 238; break;
			case 'b':
				c = 239; break;
		}
	}

	if((flags & FL_THINGS) && (sc_translate > 1))
	{
		switch(ch)
		{
		case 'r':
			c = PT_SCROLL; break;
		case 'q':
			c = PT_POTION; break;
		case '>':
			c = PT_STAIR_DOWN; break;
		case '<':
			c = PT_STAIR_UP; break;
		case 'z':
			c = PT_WAND; break;

		case 'w':
			c = PT_SWORD; break;
		case 'W':
			c = PT_ARMOUR; break;

		case 'Z':
			c = 240; break;
		case 'x':
			c = 241; break;

		case 'f':
			c = 242; break;
		//case 'X':
		//	c = 243; break;
		}
	}

	return c;
}

#define SC_X 162
#ifdef DS
#define SC_Y 95
#else
#define SC_Y 95+192
#endif

void screen_reset();

void clear_shortcuts()
{
#ifdef DS
	screen_set_target((void *)0x06200000, 0, 0, 0);
#else
	screen_reset();
#endif

	uint32 col = 0;
	screen_setcolors(&col, 0, 1);

	screen_set(SC_X, SC_Y, 40, 26, 0);
	display_flush();


}

void render_shortcuts()
{

	if(get_window() == MAP_WIN)
		return;

	Font *f = font_main;

#ifdef DS
	screen_set_target((void *)0x06200000, 0, 0, 0);
#else
	screen_reset();
#endif

	char *d;
	if(menu_mode)
	{
		font_setcolor(0xffffff, 0x0000a0, f);
		font_putchar(f, ' ',  SC_X + 13, SC_Y);
		font_putchar(f, 245,  SC_X, SC_Y + 6);
		font_putchar(f, 243,  SC_X + 26, SC_Y + 6);
		font_putchar(f, 244,  SC_X + 13, SC_Y + 13);
	}
	else
	if(locked)
	{
		d = locked;
		font_setcolor(0xffffff, 0x0000a0, f);
		font_putchar(f, translate(d[2], FL_LOCK),  SC_X + 13, SC_Y);
		font_putchar(f, translate(d[3], FL_LOCK),  SC_X, SC_Y + 6);
		font_putchar(f, translate(d[0], FL_LOCK),  SC_X + 26, SC_Y + 6);
		font_putchar(f, translate(d[1], FL_LOCK),  SC_X + 13, SC_Y + 13);
	}
	else
	{
		d = shortcuts[dirpad];
		font_setcolor(0xffffff, dirpad & 1 ? 0xff4040 : 0x0000a0, f);
		font_putchar(f, translate(d[2], FL_THINGS), SC_X + 13, SC_Y);

		font_setcolor(0xffffff, dirpad & 4 ? 0xff4040 : 0x0000a0, f);
		font_putchar(f, translate(d[3], FL_THINGS), SC_X, SC_Y + 6);

		font_setcolor(0xffffff, dirpad & 8 ? 0xff4040 : 0x0000a0, f);
		font_putchar(f, translate(d[0], FL_DIRS), SC_X + 26, SC_Y + 6);

		font_setcolor(0xffffff, dirpad & 2 ? 0xff4040 : 0x0000a0, f);
		font_putchar(f, translate(d[1], FL_THINGS), SC_X + 13, SC_Y + 13);
	}

	display_flush();
}

void lock_shortcuts(char *keys)
{
	locked = keys;
	render_shortcuts();
}


char *locktypes[] = { "+-t.", "+-q." };


void set_direction_mode(int n, int lockt)
{
	dir_mode = n;
	//if(!n)
	//	render_shortcuts();

	if(n && lockt >= 0)
		locked = locktypes[lockt];
	else
		locked = NULL;

	render_shortcuts();
}

static int save_x = -1;
static int save_y = -1;

void find_last_char()
{
	int nl = pogo_ioctl(1, CC_GETHEIGHT);

	for(int y=0; y<nl; y++)
	{
		pogo_ioctl(1, CC_GOTOXY, 0, y);
		int c = pogo_ioctl(1, CC_GETCHAR);
		if(c == last_char)
			my = y;
	}
}

/*
char menu_idx[60];
char menu_pos[60];
int menu_count = -1;

void create_menu_index()
{
	for(int i=0, int y=0; y<nl; y++)
	{
		pogo_ioctl(1, CC_GOTOXY, 0, y);
		c = pogo_ioctl(1, CC_GETCHAR);
		if(c >= 0x30 && c <= 0x7f)
		{
			menu_idx[i*2] = c;
			menu_pos[i*2] = y;

			if(c == last_picked)
				current_item = i*2;
		}

		if(menu_x2 > 0)
		{
			pogo_ioctl(1, CC_GOTOXY, menu_x2, y);
			c = pogo_ioctl(1, CC_GETCHAR);
			if(c >= 0x30 && c <= 0x7f)
			{
				menu_idx[i*2+1] = c;
				menu_pos[i*2+1] = y;
			}
			else
				menu_idx[i*2+1] = 0;

			if(c == last_picked)
				current_item = i*2+1;
		}
	}

	menu_count = i;
}
*/
void set_menu_mode(int n, int miny, int maxy, int x2)
{
	menu_mode = n;

	//if(my >= 0)
	//	console_mark_square(0, 0);

	if(n)
	{
		save_x = wherex();
		save_y = wherey();
		//create_menu_index();
	}
	else
		gotoxy(save_x, save_y);

	menu_miny = miny-1;
	menu_maxy = maxy-1;

	if(menu_miny < 0)
		menu_miny = 0;
	if(menu_maxy < 0)
		menu_maxy = 23;


	menu_x2 = x2;

	mx = 0;
	my = -1;

	//my = -1;
/*
	if(n == 1)
	{
		my = 0;
		console_mark_square(0, 0);
		display_flush();
	}*/

//	if(n)
//		clear_shortcuts();
//	else
		render_shortcuts();
		display_flush();

}

int pogo_getch(void)
{
	int c;

	if(menu_mode)
	{
		if(my == -1 && menu_mode == 2)
			find_last_char();
		if(my == -1)
			my = menu_miny;

		console_mark_square(mx, my);
	}

	display_flush();
/*
	int os = shift_pressed;
	shift_pressed |= (key_get_modifiers() & 1);
	if(os != shift_pressed)
	{
		activate_clickarea(0, shift_pressed^1);
		activate_clickarea(1, shift_pressed);
		render_panel();
	}
*/


	bool go = true;
	while(go)
	{
#ifdef DS
		swiWaitForIRQ();
#endif

		c = pogo_getchar();

		if((c == POGO_EOF) && a_key > 0 && (pogo_clock() - a_key) > key_rate)
		{
			a_key = pogo_clock();
			c = RAWKEY_A;
#ifdef DS
			key_rate = 6;
#else
			key_rate = 10;
#endif
		}

		//if(menu_mode && my < 0)
		//{
		//	my = menu_miny;
		//	console_mark_square(0, my);
		//	display_flush();
		//}

		if(c == 17)
		{
			shift_pressed ^= 1;
			//activate_clickarea(0, shift_pressed^1);
			//activate_clickarea(1, shift_pressed);
			render_panel();
		}
		else
		if(c > 0)
		{
			go = false;

			int sp = shift_pressed;

			int oc = c&0x7f;
			int od = dirpad;

			if(c == RAWKEY_A && (a_key == -1))
			{
				a_key = pogo_clock();
				key_rate = 20;
			}
			else
			if(c == (RAWKEY_A | 0x80))
			{
				a_key = -1;
			}
			else
			if(c == RAWKEY_L)
			{
				shift_pressed |= 1;
				go = true;
			}
			else
			if(c == (RAWKEY_L | 0x80))
			{
				shift_pressed &= ~1;
				go = true;
			}
			else
			if(c == RAWKEY_R)
			{
				shift_pressed |= 2;
				go = true;
			}
			else
			if(c == (RAWKEY_R | 0x80))
			{
				shift_pressed &= ~2;
				go = true;
			}
			else
			if(oc >= RAWKEY_UP && oc <= RAWKEY_RIGHT)
			{
				if(menu_mode)
				{
					console_mark_square(mx, my);

					if(c == RAWKEY_RIGHT && menu_x2 > 0)
						mx = menu_x2;
					else
					if(c == RAWKEY_LEFT && menu_x2 > 0)
						mx = 0;
					else
					if(c == RAWKEY_UP)
						my--;
					else
					if(c == RAWKEY_DOWN)
						my++;

					if(my < menu_miny) my = menu_miny;
					if(my > menu_maxy) my = menu_maxy;

					console_mark_square(mx, my);

					display_flush();
				}

				go = true;

				if(c & 0x80)
				{
					dirpad &= ~(1<<(oc-RAWKEY_UP));

					if(od && !dirpad && (dir_mode || shift_pressed) && !move)
					{
						//char shortcuts[5] = "kjhl";
						//c = shortcuts[oc-RAWKEY_UP];
						c = shortcuts[or_dirpad][0];
						or_dirpad = 0;
						go = false;
					}
				}
				else
				{
					dirpad |= (1<<(oc-RAWKEY_UP));
					or_dirpad  |= (1<<(oc-RAWKEY_UP));
				}

				if(!dirpad)
				{
					or_dirpad = 0;
					move = 0;
				}

				if(!menu_mode)
					render_shortcuts();

			}
			else
			if(menu_mode && c == RAWKEY_X)
			{

				console_mark_square(mx, my);

				int nl = pogo_ioctl(1, CC_GETHEIGHT);
				for(; my<nl; my++)
				{
					pogo_ioctl(1, CC_GOTOXY, 0, my);
					int c = pogo_ioctl(1, CC_GETCHAR);
					if(c == ' ')
					{
						while(c == ' ' && my < nl)
						{
							my++;
							pogo_ioctl(1, CC_GOTOXY, 0, my);
							c = pogo_ioctl(1, CC_GETCHAR);
						}

						break;

					}
				}

				if(my >= nl)
					my = menu_miny;

/*				if(my == menu_maxy)
					my = menu_miny;
				else
					my = menu_maxy;*/
				go = true;
				console_mark_square(mx, my);
				display_flush();
			}

			if(c & 0x80)
				go = true;


			if(shift_pressed != sp)
			{
				//activate_clickarea(0, (shift_pressed^1)&1);
				//activate_clickarea(1, shift_pressed&1);
				render_panel();
			}

		}
	}

	if(menu_mode)
	{
		console_mark_square(mx, my);
		display_flush();
	}

	int s = shift_pressed;

	switch(c)
	{
		case RAWKEY_A:

			if(menu_mode)
			{
				pogo_ioctl(1, CC_GOTOXY, mx, my);
				c = pogo_ioctl(1, CC_GETCHAR);
				if(menu_mode == 2 && (c >= 0x30 && c <= 0x7f))
					last_char = c;
				//console_mark_square(mx, my);
				//my = -1;
				//display_flush();

			}
			else
			{
				if(locked)
					c = locked[0];
				else
					c = shortcuts[dirpad][0];
				if(dirpad)
					move++;
			}
			break;
		case RAWKEY_B:
			if(menu_mode)
			{
				c = '\e';
				//console_mark_square(mx, my);
				//my = -1;
				//display_flush();
			}
			else
			{
				if(locked)
					c = locked[1];
				else
					c = shortcuts[dirpad][1];

				if(dirpad)
					move++;
			}
			break;
		case RAWKEY_X:
			if(menu_mode)
			{
				//console_mark_square(mx, my);

			//	if(my == menu_maxy)
				//	my = menu_miny;
				//else
					//my = menu_maxy;

				/*c = 0;
				while(c != ' ' && my < menu_maxy)
				{
					my++;
					pogo_ioctl(1, CC_GOTOXY, mx, my);
					c = pogo_ioctl(1, CC_GETCHAR);
				}
				if(c == ' ')
					my++;

				printf("%d %d\n", my,  menu_maxy);

				if(my > menu_maxy) my = menu_maxy;
*/
				//console_mark_square(mx, my);

				display_flush();
			}
			else
			{
				if(locked)
					c = locked[2];
				else
					c = shortcuts[dirpad][2];
				if(dirpad)
					move++;
			}
			break;
		case RAWKEY_Y:
			if(menu_mode)
			{
				c = ' ';//last_char;
				//console_mark_square(mx, my);
				my = -1;
				display_flush();
			}
			else
			{
				if(locked)
					c = locked[3];
				else
					c = shortcuts[dirpad][3];
				if(dirpad)
					move++;
			}
			break;
		case RAWKEY_SELECT:
			if(s)
				c = '\e';
			else
			if(locked)
				c = '?';
			else
				c = 'i';
			s = 0;
			break;
		case RAWKEY_START:
			if(s)
			{
				if(dirpad == 5)
					c = '&';
				else
					c = 'Q';
			}
			else
				c = 'X';
			s = 0;
			break;
	}

	if(!menu_mode && !locked)
	{
		 if(s & 1)
		{
			if(c >= 'a' && c <= 'z')
				c -= 0x20;
			else
			if(c == '.')
				c = CMD_REST;
		}
		else
		 if(s & 2)
		{
			if(c >= 'a' && c <= 'z')
				c -= 0x60;
			else
			if(c == '.')
				c = 'p' - 0x40;
		}
	}

	//if(c == 'c')
	//{
	//	volatile int x = 0;
	//}

	return c;
}


int kbhit(void)
{
//#ifdef DS
//	int keys = ((~(REG_KEYINPUT)) & 0xff);
//	return (keys != 0);
//#else
	return key_peek();
//#endif
	//return 0;
}

int key_to_command(int c)
{
	return c;
}

void get_input_line_from_curses( char *const buff, int len )
{
	//strcpy(buff, "Kazar\n");
	//pogo_gets(buff);
	readline_start(200);
	readline_set("");

	bool go_on = true;

//	int x = wherex();
//	int y = wherey();

	display_flush();

	while(go_on)
	{
#ifdef DS
		swiWaitForIRQ();
#endif
		int c = pogo_getchar();

		if((c == 17) || (c&0x7f == RAWKEY_L))
		{
			shift_pressed ^= 1;
			//activate_clickarea(0, shift_pressed^1);
			//activate_clickarea(1, shift_pressed);
			render_panel();
			continue;
		}

		if(c > 0)
		{
			int ch = 0;
			switch(c)
			{
				case RAWKEY_RIGHT:
					ch = RL_KEY_RIGHT;
					break;
				case RAWKEY_LEFT:
					ch = RL_KEY_LEFT;
					break;
				case RAWKEY_UP:
					ch = RL_KEY_HOME;
					break;
				case RAWKEY_DOWN:
					ch = RL_KEY_END;
					break;
				case 8:
					ch = RL_KEY_DEL;
					break;
				case 13:
				case 10:
				{
					strcpy(buff, readline_end(1));
					return;
				}
				default:
					if(c >= 0x20 && c <= 0x7f)
						ch = c;
			}
			if(ch >= 0)
			{
				if(readline_putchar(ch))
					display_flush();
			}
		}
	}


    //readline_set(filename);
   //	malloc(4096);


}
