#include "AppHdr.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <ctype.h>

#ifdef _WIN32
#include <io.h>
#include <windows.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#include <pogo/pogo.h>
#include <pogo/console.h>
#include <pogo/keyboard.h>
#include <pogo/readline.h>

#include "zlib.h"

#include "bmp.h"

#ifdef DS
#include <nds.h>
#include <gbfs.h>
#include <nds/bios.h>
extern GBFS_FILE *gbfs_root;
#endif

#include "defines.h"
#include "enum.h"
#include "tags.h"


extern "C" {
	void sdlkey_init();
	void key_init();
	void screen_flush();
	int key_peek();
	int fatdev_init();
	void *gbfsdev_init();
	void add_clickarea(int x, int y, int w, int h, int charw, uchar *);
	void activate_clickarea(int, int);
	int key_get_modifiers();
	void console_mark_square(int x, int y);

}

void render_shortcuts();

void modplay_init();
void modplay_playsong(int n);


/////////////////////////////////

extern int shift_pressed;


uint32 panel_cols[256];
uchar *panel_ptr;
int panel_w, panel_h;

Font *font_main, *font_map;


static int cursor = 1;

int character_set;
int winn = 1;

int wins[8];
int did_draw[8];
int marked_x = -1;
int marked_y;


unsigned char buffer[65536];

void pogo_pack(char *data, tagHeader *th)
{
	int ret;
	z_stream strm;

	strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;
    ret = deflateInit(&strm, Z_DEFAULT_COMPRESSION);

    if(ret != Z_OK)
    {
     	cprintf("PACK FAIL! %d\n", ret);
    	while(1);
    }

	strm.next_in = (Bytef*)data;
	strm.avail_in = th->offset;
	strm.avail_out = 65536;
	strm.next_out = buffer;
	ret = deflate(&strm, Z_FINISH);

    if(ret != Z_OK && ret != Z_STREAM_END)
    {
     	cprintf("PACK FAIL! %d\n", ret);
		while(1);
    }

	int size = strm.total_out;

	data[0] = -65;
	data[1] = -5;

	memcpy(data+2, buffer, size);
	th->offset = size+2;

	deflateEnd(&strm);
}

void pogo_unpack(char *data, tagHeader *th)
{
	int ret;
	z_stream strm;

	if(data[0] != -65 || data[1] != -5)
	{
		cprintf("ERROR\n");
		while(1);
	}

	strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;
    ret = inflateInit(&strm);

	if(ret != Z_OK)
    {
    	cprintf("UNPACK FAIL! %d\n", ret);
    	while(1);
    }

	strm.next_in = (Bytef*)data+2;
	strm.avail_in = th->offset-2;
	strm.avail_out = 65536;
	strm.next_out = buffer;
	ret = inflate(&strm, Z_FINISH);

	if(ret != Z_OK && ret != Z_STREAM_END)
    {
    	cprintf("UNPACK FAIL! %d\n", ret);
    	while(1);
    }

	int size = strm.total_out;

	memcpy(data, buffer, size);
	th->offset = size;

	inflateEnd(&strm);
}


void screen_reset()
{
	Pixel *ptr;
	display_get_target(&ptr, NULL, NULL);
	screen_set_target(ptr, 0,  0);
}

void render_panel()
{
	if(winn == MAP_WIN)
		return;
	if(panel_ptr)
	{
		//int s = panel_h * panel_w / 2;

		//screen_setcolors(panel_cols, 0, 256);
#if DS
		//screen_set_target((void *)0x06200000, 256, 192, 0);
		//screen_copy(165, 5, panel_ptr + (shift_pressed&1)*s, panel_w, panel_h/2, panel_w, 1);
#else
		//screen_reset();
		//screen_copy(165, 5+192, panel_ptr + (shift_pressed&1)*s, panel_w, panel_h/2, panel_w, 1);
#endif
		display_flush();
	}
}

//////



PogoDevice nulld;

int null_open(const char *name, int flags)
{
	if(strlen(name) != 0)
		return -1;
	return 0;
}

int null_read(int fd, void *dest, int size)
{
	return 0;
}


int null_write(int fd, const void *dest, int size)
{
	return 0;
}


void null_close(int fd)
{
}

void null_init()
{
	memset(&nulld, 0, sizeof(PogoDevice));
	nulld.open = null_open;
	nulld.read = null_read;
	nulld.write = null_write;
	nulld.close = null_close;
	pogo_device_register(&nulld, "/null", NULL, -1);
}



PogoDevice native;

static FILE *open_files[32];
static int init = 0;


int native_open(const char *name, int flags)
{
	FILE *f;

	if(!init)
	{
		memset(open_files, 0, sizeof(FILE*) * 32);
		init = 1;
	}

	if(flags & PGO_WRONLY)
		f = fopen(name, "wb");
	else
		f = fopen(name, "rb");
	if(f)
	{
		int i = 0;
		while(open_files[i]) i++;
		open_files[i] = f;
		return i;
	}
	return -1;
}

int native_read(int fd, void *dest, int size)
{
	FILE *f = open_files[fd];
	return fread(dest, 1, size, f);
}

int native_write(int fd, const void *dest, int size)
{
	FILE *f = open_files[fd];
	return fwrite(dest, 1, size, f);
}

void native_close(int fd)
{
	FILE *f = open_files[fd];
	fclose(f);
	open_files[fd] = NULL;
}

int native_lseek(int fd, int offset, int origin)
{
	FILE *f = open_files[fd];
	fseek(f, offset, origin);
	return ftell(f);
}

int native_tell(int fd)
{
	FILE *f = open_files[fd];
	return ftell(f);
}

int native_ioctl(int fd, int request, ...)
{
	return -1;
}

int native_remove(const char *name)
{
	return -1;
}

#ifdef _WIN32

static struct _finddata_t fdata;
static intptr_t handle;

int native_dopen(const char *name)
{
	char tmp[256];
	sprintf(tmp, "%s/*", name);
	handle = _findfirst(tmp, &fdata);
	return handle;
}

int native_dread(int fd, void *dest, int len)
{
	int rc;

	if(handle < 0)
		return 0;

	strcpy((char*)dest, fdata.name);
	rc = _findnext(handle, &fdata);
	if(rc == -1)
		handle = -1;
	return strlen((char*)dest);
}

void native_dclose(int fd)
{
	_findclose(handle);
	//return 0;
}

#else

int native_dopen(const char *name)
{
	return -1;
}

int native_dread(int fd, void *dest, int len)
{
	return strlen((char*)dest);
}

void native_dclose(int fd)
{
	//return 0;
}


#endif

int native_stat(const char *name, struct pogo_stat *buffer)
{
	struct stat s;
	int rc = stat(name, &s);
	if(rc == 0)
	{
		buffer->pst_size = s.st_size;
		buffer->pst_atime = (uint32)s.st_atime;
		if(s.st_mode & S_IFDIR)
			buffer->pst_mode = PGS_IFDIR;
		else
			buffer->pst_mode = 0;
	}

	return rc;
}



static void native_init()
{
	memset(&native, 0, sizeof(PogoDevice));
	native.open = native_open;
	native.read = native_read;
	native.write = native_write;
	native.close = native_close;
	native.lseek = native_lseek;
	native.stat = native_stat;
	native.dopen = native_dopen;
	native.dread = native_dread;
	native.dclose = native_dclose;

	pogo_device_register(&native, "/disk", NULL, -1);
}


void mark_square(int x, int y)
{
	int redraw = 0;

	pogo_ioctl(1, CC_SETWIN, wins[PLAY_WIN]);

	if(marked_x >= 0)
	{
		console_mark_square(marked_x-1, marked_y-1);
		redraw = 1;
	}

	marked_x = x; marked_y=y;
	if(x >=0)
	{
		console_mark_square(x-1, y-1);
		redraw = 1;
	}

	if(redraw)
		display_flush();

	pogo_ioctl(1, CC_SETWIN, wins[winn]);


}

void update_cursor()
{
}

int clrscr(void)
{
	pogo_ioctl(1, CC_CLEAR);
	return 0;
}

int cprintf(const char *format,...)
{
	char tmp[2560];
	int rc;
	va_list vl;
	va_start(vl, format);

	did_draw[winn] = 1;

	mark_square(-1, -1);

	rc = vsprintf(tmp, format, vl);
	pogo_write(1, tmp, strlen(tmp));

	update_cursor();

	va_end(vl);
	return rc;

	return 0;
}

extern "C" {
int cprintf2(const char *format,...)
{
	char tmp[2560];
	int rc;
	va_list vl;
	va_start(vl, format);

	rc = vsprintf(tmp, format, vl);
	pogo_write(1, tmp, strlen(tmp));

	update_cursor();

	va_end(vl);
	return rc;
}
}


int gotoxy(int x, int y)
{
	//x--;
	//y--;
	//if(x < 0) x = 0;
	//if(y < 0) y = 0;
	pogo_ioctl(1, CC_GOTOXY, x-1, y-1);
	update_cursor();
	return 0;
}


int putch(unsigned char c)
{
	did_draw[winn] = 1;

	mark_square(-1, -1);

	if(c < 32)
	{
		c = ' ';
	}
	else
	if(winn == MAP_WIN )
	{
		int t = LIGHTGREEN;
		if((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
			t = CYAN;
		else
		switch(c)
		{
			case PT_HUMAN:
				t = WHITE;
				break;
			case PT_DOOR:
				t = LIGHTGRAY;
			case PT_OPEN_DOOR:
				break;
			case PT_FLOOR:
				t = DARKGRAY;
				break;
			case PT_STAIR_DOWN:
				t = LIGHTRED;
				break;
			case PT_STAIR_UP:
				t = LIGHTBLUE;
				break;
			case PT_TRAP:
				t = CYAN;
				break;
			case PT_STONE_WALL:
			case PT_BRICK_WALL:
			case PT_BRICK_WALL2:
			case PT_BLOCK_WALL:
			case PT_STONE_WALL2:
				t = BROWN;
				break;
			case PT_WATER:
			case PT_WATER2:
				t = BLUE;
				break;
			case PT_ENTRANCE:
			case PT_HOUSE:
				t = LIGHTYELLOW;
				break;
		}
		if(t >= 0)
			textcolor(t);
	}
	else
	if(winn == STAT_WIN)
	{
		switch(c)
		{
		case 249:
			c = '.'; break;
		case 177:
			c = '#'; break;
		}
	}
//	else
//	if(winn == PLAY_WIN)
//	{
//		if(c == 144)
//			textcolor(17);
//	}

	pogo_putchar(c);
	update_cursor();
	return 0;
}

int translate_keypad(int keyin)
{
	return 0;
}

int wherex(void)
{
	return pogo_ioctl(1, CC_GETX) + 1;
}

int wherey(void)
{
	return pogo_ioctl(1, CC_GETY) + 1;
}

int window(int x1, int y1, int x2, int y2)
{
	return 0;
}



void update_screen(void)
{
	display_flush();
}

static char spaces[128] = "";

void clear_to_end_of_line(void)
{
	if(!*spaces)
		memset(spaces, 0x20, 128);

	int w = pogo_ioctl(1, CC_GETWIDTH);
	int x = pogo_ioctl(1, CC_GETX);
	int y = pogo_ioctl(1, CC_GETY);
	pogo_write(1, spaces, w-x);
	pogo_ioctl(1, CC_GOTOXY, x, y);
}

void clear_to_end_of_screen(void)
{
	if(!*spaces)
		memset(spaces, 0x20, 128);

	int w = pogo_ioctl(1, CC_GETWIDTH);
	int h = pogo_ioctl(1, CC_GETHEIGHT);
	int y = pogo_ioctl(1, CC_GETY);

	clear_to_end_of_line();

	for(int i=0; i<h-y; i++)
	{
		pogo_write(1, spaces, w);
		pogo_putchar(10);
	}
}

int get_number_of_lines_from_curses(void)
{
	return pogo_ioctl(1, CC_GETHEIGHT);
	//return 0;
}


void _setcursortype(int curstype)
{
	cursor = curstype;
	update_cursor();
}

void delay(unsigned long t)
{
	//int w = pogo_time(NULL) + t/10;
	//while(w > pogo_time(NULL))
	//{
#ifdef DS
	swiWaitForIRQ();
	swiWaitForIRQ();
#endif
	//}
	display_flush();
}

void init_key_to_command()
{
}

int cbg = -1;
int cfg = -1;

void textbackground(int col)
{
	if(cbg != col)
	{
		cbg = col;
		pogo_ioctl(1, CC_SETCOLOR, cfg, cbg);
	}
}

void textcolor(int col)
{
	if(cfg != col)
	{
		cfg = col;
		pogo_ioctl(1, CC_SETCOLOR, cfg, cbg);
	}
}



Font *bmp2font(char *fname, int cw, int ch, int first, int last)
{
	void *ptr = NULL;
	int w,h;
	int rc = read_bmp(fname, &ptr, 8, &w, &h, 0, 0);

	if(rc == ERR_BMP_OK)
	{
		Font *f = (Font *)malloc(sizeof(Font));
		memset(f, 0, sizeof(Font));
		f->flags = 0;
		f->first = first;
		f->last = last;
		f->charwidth = cw;
		f->height = ch;//h;
		f->width = w;//10*224;
		f->name = "10x10";
		f->pixels = (uchar *)ptr;

		//cprintf("Made font %s from %x\n", fname, ptr);

		return f;
	}
	//else
	//	cprintf("%s failed (%d)\n", fname, rc);

	return NULL;
}


#ifdef DS


bool test_sram()
{
	bool rs_writeable = false;
	bool rs_8bit = false;

	//WAIT_CR = 0x6003;

	volatile char *sram = (volatile char *)0x0A000000;
	volatile uint16 *sram16 = (volatile uint16*)sram;

	volatile char c0 = *sram;

	//cprintf("Read %02x, wrote %02x\n", c0, c0 ^ 0xff);
	*sram = (c0 ^ 0xff);
	//cprintf("Read back %02x\n", *sram);

	if(*sram != c0)
	{
		*sram = c0;
		rs_writeable = true;

		c0 = sram[0];
		char c1 = sram[1];
		uint16 c2 = *sram16;
		*sram16 = (c2 ^ 0x55AA);

		if(*sram != (c2 ^ 0x55AA))
			rs_8bit = true;

		sram[0] = c0;
		sram[1] = c1;
	}

	//WAIT_CR = 0xE880;

	return rs_writeable && rs_8bit;

}
//cprintf("GBA SRAM is %s\n", rs_writeable ? (rs_8bit ? "8bit writeable" : "not 8bit (RAM?)") : "not present");
#endif

int check_file(char *name, int minsize = 1)
{
	POGOFILE *fp = pogo_fopen(name, "rb");
	if(fp)
	{
		pogo_fseek(fp, 0, SEEK_END);
		int l = pogo_ftell(fp);
		pogo_fclose(fp);
		if(l >= minsize)
			return 1;
	}

	return 0;
}

int check_write(char *name)
{
	int rc = 0;
	int l;
	char dummy[32];

	strcpy(dummy, "TestContent");

	POGOFILE *fp = pogo_fopen(name, "wb");
	if(fp)
	{
		rc++;
		l = pogo_fwrite(dummy, 25, 1, fp);
		if(l == 25)
			rc++;
		pogo_fclose(fp);

		fp = pogo_fopen(name, "rb");
		if(fp)
		{
			rc++;
			memset(dummy, 0, 32);
			l = pogo_fread(dummy, 32, 1, fp);
			if(l > 0)
				rc++;
			if(l == 25)
			{
				rc++;
				if(strcmp(dummy, "TestContent") == 0)
					rc++;
			}
		}
		pogo_fclose(fp);

		pogo_remove(name);
	}

	return rc;
}

extern unsigned char colors[256][3];

extern "C" {
	void rdev_init(void *mem, int size);
	void test_fs();
	void rinit(void *memory, int size, int format);

}

int use_gbacart = 0;

static char *savedir;
static char tempfilename[128];

char *SAVEFILENAME(char *x)
{
	sprintf(tempfilename, "%s%s", savedir, x);
	//if(!use_gbacart)
//		return x;
//	strcpy(&tmpfilename[5], x);
	return tempfilename;
}

void rinit(void *memory, int size, int format);

const char *kdef[] =
{
	"aA2bB2cC2dD2eE2fF2gG2",
	"hH2iI2jJ2kK2lL2mM2nN2",
	"oO2pP2qQ2rR2sS2tT2uU2",
	"vV2wW2xX2yY2zZ2 e2  2",
	"?=2**2.#2,/2<!2>^2 b2",
	" x2~;20@21[22\\23]24\"2",
	" s2` 25+26-27:28 2' 2",
	NULL
};


void init_libpogo()
{
	pogo_device_init();
	display_set_params(256, 192*2, 0);
	display_init();

#ifdef DS

	int fat_works = 0;
	void *gbfs_works = 0;
	int fat_write = 0;
	int fat_hasfiles = 1;

	char tmp[80];
	font_setcolor(0x000000, 0xFFFFFF, NULL);
	sprintf(tmp, "Initing fat filesystem (%s) ...", FATLIB);
	font_text(font_getdefault(), tmp, 0, 0);
	fat_works = fatdev_init();
	if(fat_works)
	{
		fat_works = 2;
		fat_write = check_write("testfile");
		if(fat_write != 6)
			use_gbacart = 1;

		fat_hasfiles = 0;
	}
	else
	{
		gbfs_works = gbfsdev_init();
		use_gbacart = 1;
	}

#else
	native_init();
#endif

	null_init();
	modplay_init();

	key_init();
	console_init();

	pogo_open("/dev/key", 0);
	pogo_open("/dev/console", 0);

	savedir = "";

	if(use_gbacart)
		savedir = "/ram/";

#ifdef DS

	if(use_gbacart)
		savedir = "/ram/";
	else
	if(fat_works)
	{
		int fd = pogo_dopen("/fat/CRAWL");
		if(fd >= 0)
		{
			savedir = "/fat/CRAWL/";
			pogo_dclose(fd);
		}
		else
			savedir = "/fat";
	}
	else
		savedir = "";

	cprintf("Initing fat filesystem (%s) ...\n", FATLIB);
	
	if(!fat_works)
	{
		cprintf("No FAT filesystem detected\n");

		if(gbfs_works)
			cprintf("Using internal (GBFS) at %08x\n", gbfs_works);
		else
		{
			cprintf("GBFS datafiles not found in ROM!\n");
			while(1);
		}
	}
	else
	if(fat_write != 6)
	{
		cprintf("FAT filesystem is not writeable (%d).\n", fat_write);
		cprintf("Data still loaded from FAT.\n");
	}
	else
		cprintf("Using FAT filesystem for loading/saving\n");

	if(use_gbacart)
	{
		if(!test_sram())
		{
			cprintf("Could not detect GBA cart SRAM!\n");
			cprintf("RAM will be used instead which means your game\n");
			cprintf("will be lost when you switch off your machine.\n");
			rdev_init(malloc(65536), 65536);
		}
		else
		{
			cprintf("Using cart SRAM for saving.\n");
			rdev_init((void *)0x0a000000, 65536);
			int ram_check = check_write("/ram/testfile");
			cprintf("SRAM filsystem is %s (%d)\n", (ram_check == 6) ? "OK" : "Not OK!?", ram_check);
		}
	}

	cprintf("Saving to '%s'\n",savedir);

	int ck = 0;

	while(ck != RAWKEY_START && ck != ' ')
	{
		cprintf("Press START...\n");
		display_flush();

		while((ck = pogo_getchar()) <= 0);

		if(ck == RAWKEY_SELECT)
		{
			struct pogo_stat sbuf;
			char name[64], tmp[64];
			int total = 0;
			int fcount = 0;
			int fd;
			clrscr();
			cprintf("SRAM Contents:\n");
			fd = pogo_dopen("/ram/");
			while(pogo_dread(fd, name, sizeof(name)) >= 0)
			{
				sprintf(tmp, "/ram/%s", name);
				pogo_stat(tmp, &sbuf);
				pogo_ioctl(1, CC_GOTOXY, (fcount&1)*25, fcount/2);
				cprintf("%s (%d bytes)", name, sbuf.pst_size);
				total += sbuf.pst_size;
				fcount++;
			}
			pogo_ioctl(1, CC_GOTOXY, 0, 22);
			cprintf("*** TOTAL: %d bytes in %d files\n", total, fcount);
			pogo_dclose(fd);
		}
		else
		if(ck == RAWKEY_X)
		{
			int up = 0;
			ck = 0;
			clrscr();
			cprintf("Really nuke SRAM? (UP,DOWN to confirm)\n");
			display_flush();
			while(1)
			{
				ck = pogo_getchar();
				if(!(ck & 0x80))
				{
					if(up == 0 && ck == RAWKEY_UP)
						up=1;
					else
					if(up == 1 && ck == RAWKEY_DOWN)
					{
						rinit(NULL, 0, 1);
						cprintf("*** NUKED!\n");
						break;
					}
					else
					{
						cprintf("...Cancelled\n");
						break;
					}
				}
			}
		}
/*		else
		if(ck == 'X')
		{
			int fd,fd2,rc;
			char buffer[65536];
			char name[64], tmp[64];
			fd = pogo_dopen("/ram/");
			while(pogo_dread(fd, name, sizeof(name)) >= 0)
			{
				sprintf(tmp, "/ram/%s", name);

				fd2 = pogo_open(tmp, 0);
				rc = pogo_read(fd2, sizeof(buffer), buffer);
				pogo_close(fd2);

				sprintf(tmp, "SAVE/%s", name);
				fd2 = pogo_open(tmp, PGO_WRONLY);
				pogo_write(fd2, rc, buffer);
				pogo_close(fd2);
			}
		}*/
	}
#endif


	Font *f;
	Font *font[4];

	font[0] = font[1] = font[2] = font[3] = font_getdefault();
	int w,h;

	screen_get_target(NULL, &w, &h, NULL);

	void *ptr;

	int rc;

	panel_ptr = NULL;

	rc = read_bmp("panel.bmp", &ptr, 8, &panel_w, &panel_h, 0, 0);

	if(rc == ERR_BMP_OK)
	{
		panel_ptr = (uchar*)ptr;
		for(int i=0; i<256; i++)
			panel_cols[i] = (colors[i][2]<<16) | (colors[i][1]<<8) | colors[i][0];
	}

	//add_clickarea(165, 5, 7*12, 7*12, 12, (uchar*)"abcdefghijklmnopqrstuvwxyz\n ?*.,<>\010\e;@[\\]\"\021\020+-:&'");
	//add_clickarea(165, 5, 7*12, 7*12, 12, (uchar*)"ABCDEFGHIJKLMNOPQRSTUVWXYZ\n{=*#/!^\010\e~01234\021`56789");
	//activate_clickarea(1, 0);
	
	
	int kh = pogo_ioctl(POGO_STDIN, KC_SETUP_KEYBOARD, 165, 192+5, w-170, kdef, font_getdefault());



	font_main = bmp2font("12x11font.bmp", 12, 11, 0x20, 255);
	if(!font_main)
		font_main = font_getdefault();
	font_map = 	bmp2font("3x3mapfont.bmp", 3, 3, 0x20, 255);
	if(!font_map)
		font_map = font_getdefault();

#ifdef DS

	font[0] = font[1] = font[2] = font[3] = font_main;
	console_set_target((void *)0x06000000, w);
	wins[PLAY_WIN] = pogo_ioctl(1, CC_OPENWIN, 0, 0, w, h, font, CONWIN_FIXED);

	font[0] = font[1] = font[2] = font[3] = font_map;
	console_set_target((void *)0x06200000, w);
	wins[MAP_WIN] = pogo_ioctl(1, CC_OPENWIN, 0, 0, w, h, font, CONWIN_FIXED);

	font[0] = font[1] = font[2] = font[3] = font_getdefault();
	wins[STAT_WIN] = pogo_ioctl(1, CC_OPENWIN, 0, 0, 160, 132, font, CONWIN_FIXED);
	
	wins[MSG_WIN] = pogo_ioctl(1, CC_OPENWIN, 0, 132, 256, 60, font, CONWIN_FIXED);

	console_set_target((void *)0x06000000, 256);
	wins[TEXT_WIN] = pogo_ioctl(1, CC_OPENWIN, 0, 0, 256, 192, font, CONWIN_FIXED);

#else

	font[0] = font[1] = font[2] = font[3] = font_main;
	wins[PLAY_WIN] = pogo_ioctl(1, CC_OPENWIN, 0, 0, 256, 192, font, CONWIN_FIXED);

	font[0] = font[1] = font[2] = font[3] = font_map;
	wins[MAP_WIN] = pogo_ioctl(1, CC_OPENWIN, 0, 192, 256, 192, font, CONWIN_FIXED);

	font[0] = font[1] = font[2] = font[3] = font_getdefault();
	wins[STAT_WIN] = pogo_ioctl(1, CC_OPENWIN, 0, 192, 160, 132, font, CONWIN_FIXED);
	
	wins[MSG_WIN] = pogo_ioctl(1, CC_OPENWIN, 0, 192+132, 256, 60, font, CONWIN_FIXED);
	
	wins[TEXT_WIN] = pogo_ioctl(1, CC_OPENWIN, 0, 0, 256, 192, font, CONWIN_FIXED);

#endif

	winn = TEXT_WIN;
	if(panel_ptr)
		render_panel();

	modplay_playsong(1);

}

int get_window() { return winn; }

int set_window(int n)
{
	int ow = winn;

	if(winn == TEXT_WIN && did_draw[winn])
		clrscr();
	else
	if(ow == MAP_WIN)
		clrscr();

	winn = n;

	did_draw[winn] = 0;

	pogo_ioctl(1, CC_SETWIN, wins[n]);

	if(winn == TEXT_WIN || winn == MAP_WIN)
	{
		render_panel();
		render_shortcuts();
	}

//	if(n == PLAY_WIN && ow == TEXT_WIN)
	//	clrscr();

	return ow;

}

void lincurses_startup()
{
}

void lincurses_shutdown()
{
}

#ifndef CYGWIN
/* Convert value to string */
int itoa(int value, char *strptr, int radix)
{
    unsigned int bitmask = 32768;
    int ctr = 0;
    int startflag = 0;

    if (radix == 10)
    {
        sprintf(strptr, "%i", value);
    }
    if (radix == 2)             /* int to "binary string" */
    {
        while (bitmask)
        {
            if (value & bitmask)
            {
                startflag = 1;
                sprintf(strptr + ctr, "1");
            }
            else
            {
                if (startflag)
                    sprintf(strptr + ctr, "0");
            }

            bitmask = bitmask >> 1;
            if (startflag)
                ctr++;
        }

        if (!startflag)         /* Special case if value == 0 */
            sprintf((strptr + ctr++), "0");

        strptr[ctr] = (char) NULL;
    }
    return 1;                /* Me? Fail? Nah. */
}

char *strlwr(char *str)
{
	unsigned int i;

	for (i = 0; i < strlen(str); i++)
		str[i] = tolower(str[i]);

	return (str);
}

#endif

void rewind(POGOFILE *stream)
{
	pogo_fseek(stream, 0, PG_SEEK_SET);
}

void pogo_perror(const char *str)
{
	gotoxy(1,1);
	cprintf(str);
}


POGOFILE *pogo_smartopen(const char *name, const char *flg)
{
	char tmp[64];
	POGOFILE *fp;
	if(*flg == 'r')
	{
#ifdef DS
		sprintf(tmp, "/fat/CRAWL/%s", name);
		if(fp = pogo_fopen(tmp, flg))
			return fp;
		sprintf(tmp, "/fat/%s", name);
		if(fp = pogo_fopen(tmp, flg))
			return fp;
		sprintf(tmp, "/rom/%s", name);
		if(fp = pogo_fopen(tmp, flg))
			return fp;
#else
		sprintf(tmp, "data/%s", name);
		if(fp = pogo_fopen(tmp, flg))
			return fp;
		if(fp = pogo_fopen(name, flg))
			return fp;
#endif
	}
	else
	{
#ifdef DS
		sprintf(tmp, "/fat/CRAWL/%s", name);
		if(fp = pogo_fopen(tmp, flg))
			return fp;
		sprintf(tmp, "/fat/%s", name);
		if(fp = pogo_fopen(tmp, flg))
			return fp;
#else
		if(fp = pogo_fopen(name, flg))
			return fp;
#endif
	}

	return NULL;
}


extern "C" {
POGOFILE *pogo_smartopen2(const char *name, const char *flg)
{
	return pogo_smartopen(name, flg);
}
}


int action_hook(int action, int a0, int a1, int a2)
{
	switch(action)
	{
		case ACT_NEW_LEVEL:
			modplay_playsong(a0+3);
			break;
		case ACT_HISCORE:
			modplay_playsong(2);
			break;
	}
	return 0;
}
