#ifdef DS
#include <nds.h>
#endif

#include <pogo/types.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>


/*
 FORMAT

"MINSF5",0,0
	uint32 next;
	uint32 length;
	uint32 hsize;
	char name[0];
	...
*/

char header[128];
uchar * volatile mem_start;
uchar * volatile mem_end;
uchar * volatile mem_ptr;

void SeekTo(int x) {
	mem_ptr = &mem_start[x];
}

void Seek(int x) {
	mem_ptr += x;
}

void ReadBytes(uchar *dest, int size) {
#ifdef DS
	//WAIT_CR = 6000;
#endif
	while(size--)
		*dest++ = *mem_ptr++;
#ifdef DS
	//WAIT_CR = 0xE880;
#endif
}

void WriteBytes(uchar *src, int size)
{
	int i, tries = 5, fail;
#ifdef DS
	//WAIT_CR = 6000;
#endif

	do
	{
		fail = 0;
		tries--;

		for(i=0; i<size; i++)
			mem_ptr[i] = src[i];

		for(i=0; i<size; i++)
		{
			if(src[i] != mem_ptr[i])
			{
				fail = 1;
				break;
			}
		}
	}
	while(fail && tries);

	if(!tries)
	{
		while(1);
	}


	mem_ptr += size;
#ifdef DS
	//WAIT_CR = 0xE880;
#endif
}

int GetPos() { return mem_ptr - mem_start; }


int32 ReadInt32()
{
	int32 x;
	ReadBytes((uchar *)&x, sizeof(int32));
	return x;
}

void WriteInt32(int v)
{
	int32 x = v;
	WriteBytes((uchar *)&x, sizeof(int32));
}

///////////////////////////

int32 first_file()
{
	int pos;
	SeekTo(8);
	pos = ReadInt32();
	if(pos)
	{
		SeekTo(pos);
		return pos;
	}

	return 0;
}

int32 next_file(int pos)
{
	SeekTo(pos);
	int32 next = ReadInt32();
	if(next)
	{
		SeekTo(next);
		return next;
	}
	return 0;
}

int32 last_file()
{
	int32 next;
	int32 current = first_file();

	if(!current)
		return 0;

	do
	{
		next = next_file(current);
		if(next)
			current = next;
	}
	while(next);

	SeekTo(current);

	return current;
}

int32 get_fileinfo(int offs, int32 *length, char *filename)
{
	uchar buffer[128];
	int32 n,l,h;

	SeekTo(offs);

	n = ReadInt32();
	l = ReadInt32();
	h = ReadInt32();

	if(length) *length = l;

	ReadBytes(buffer, h);
	if(filename)
		strcpy(filename, buffer);
	return GetPos();
}


// Find the offset of a file, -1 if not found
int32 find_file(const char *fname)
{
	int next, length, hsize, pos;

	next = first_file();

	while(next)
	{
		SeekTo(next);

		pos = GetPos();

		next = ReadInt32();
		length = ReadInt32();
		hsize = ReadInt32();
		ReadBytes(header, hsize);

		if(strcmp(fname, header) == 0)
		{
			SeekTo(pos);
			return pos;
		}

	}

	return 0;
}

void dump_files()
{
	int next, length, hsize, pos;

	next = first_file();

	while(next)
	{
		SeekTo(next);
		pos = GetPos();
		next = ReadInt32();
		length = ReadInt32();
		hsize = ReadInt32();
		ReadBytes(header, hsize);

		printf("[%08x -%08x] \"%s\" (%d bytes)\n", pos, pos+12+hsize+length, header, length);

	};
	printf("\n");

}

static uchar buffer[65536];

int compact_files_forward(int end_file)
{
	int next_endpos, hsize, length, current = 0;
	int startpos, endpos = 12;
	int next = first_file();

	if(next)
	{
		SeekTo(8);
		WriteInt32(12);
	}

	while(next)
	{
		SeekTo(next);

		current = next;

		startpos = GetPos();

		next = ReadInt32();
		length = ReadInt32();
		hsize = ReadInt32();

		next_endpos = endpos + hsize + length + 12;

		if(next == next_endpos)
		{
			if(current == end_file)
				return endpos;
			endpos = next_endpos;
			continue;
		}

		ReadBytes(buffer, hsize+length);
		SeekTo(endpos);

		if(current == end_file)
			WriteInt32(next);
		else
			WriteInt32(next ? endpos + hsize + length + 12 : 0);


		WriteInt32(length);
		WriteInt32(hsize);
		WriteBytes(buffer, hsize+length);

		if(current == end_file)
			return endpos;

		endpos = GetPos();
	}

	return 0;
}



void compact_files_backward(int end_file)
{
	int hsize, length;
	int i, filecount;
	int32 file_offsets[256];


	int next = first_file();

	filecount = 0;

	while(next)
	{
		file_offsets[filecount++] = next;
		SeekTo(next);
		next = ReadInt32();
	}

	int file_pos = 0;

	for(i=filecount-1; i>=0; i--)
	{
		SeekTo(file_offsets[i]);

		if(file_offsets[i] == end_file)
		{
			WriteInt32(file_pos);
			return;
		}

		next = ReadInt32();


		length = ReadInt32();
		hsize = ReadInt32();

		if(file_pos)
			next = file_pos;

		file_pos = (file_pos ? file_pos : mem_end - mem_start) - 12 - hsize - length;

		if(file_pos == file_offsets[i])
			continue;

		ReadBytes(buffer, hsize+length);


		SeekTo(file_pos);

		WriteInt32(next);
		WriteInt32(length);
		WriteInt32(hsize);
		WriteBytes(buffer, hsize+length);
	}

}

void write_file(int file_pos, int offs, uchar *src, int size)
{
	int32 length, hsize;
	SeekTo(file_pos+4);
	length = ReadInt32();
	hsize = ReadInt32();

	Seek(hsize+offs);

	WriteBytes(src, size);

	if(offs+size > length)
	{
		length = offs+size;
		SeekTo(file_pos+4);
		WriteInt32(length);
	}
}

void truncate_file(int file_pos, int size)
{
	SeekTo(file_pos+4);
	WriteInt32(size);
}


int  bump_file(int offs)
{
	int rc;
	compact_files_backward(offs);
	rc = compact_files_forward(offs);

	//printf("Bumped %08x to %08x\n", offs, rc);
	dump_files();

	return rc;
}



void read_file(int file_pos, int offs, uchar *dst, int size)
{
	int32 hsize;
	SeekTo(file_pos+8);
	hsize = ReadInt32();

	Seek(hsize+offs);

	ReadBytes(dst, size);
}


void remove_file(int offs)
{
	int next, pos;

	next = first_file();

	if(next == offs)
	{
		// Removing first file
		SeekTo(next);
		next = ReadInt32();

		SeekTo(8);
		WriteInt32(next);
		return;

	}

	while(next)
	{
		SeekTo(next);
		pos = GetPos();
		next = ReadInt32();

		if(next == offs)
		{
			SeekTo(next);
			next = ReadInt32();
			SeekTo(pos);
			WriteInt32(next);
			return;
		}
	}
}


int create_file(const char *fname)
{
	int next, length, hsize, pos, l;

	pos = last_file();
	l = strlen(fname);

	if(pos)
	{
		compact_files_forward(pos);
		pos = last_file();

		next = ReadInt32();
		length = ReadInt32();
		hsize = ReadInt32();

		//assert(!next);

		next = pos + hsize + length + 12;
		Seek(-12);
		WriteInt32(next);
		Seek(8);
		SeekTo(next);
	}
	else
	{
		SeekTo(8);
		WriteInt32(12);
	}

	pos = GetPos();
	WriteInt32(0);
	WriteInt32(0);
	WriteInt32(l+1);
	WriteBytes((uchar*)fname, l+1);

	return pos;
}



//int cprintf2(const char *format,...);

void rinit(void *memory, int size, int format)
{
	char dest[12];

	if(memory)
	{
		mem_start = (char *)memory;
		mem_end = &mem_start[size];
	}

	SeekTo(0);
	ReadBytes(dest, 12);

	if(format || strcmp(dest, "MINSF5") != 0)
	{
		//cprintf2("Creating new\n");

		memset(dest, 0, 12);
		strcpy(dest, "MINSF5");
		SeekTo(0);
		WriteBytes(dest, 12);
	}
	//else
	//{
	//	cprintf2("Found existing\n");
	//}

}

/*
	int offs;
	uchar data[1024];
	void *ptr = malloc(70000);
	rinit(ptr+1, 65536);

	offs = create_file("firstfile");
	write_file(offs, 0, data, 1365);

	offs = create_file("remove1");
	write_file(offs, 0, data, 580);

	offs = create_file("file1");
	write_file(offs, 0, data, 765);

	offs = create_file("foobar");
	write_file(offs, 0, data, 362);

	offs = create_file("remove2");
	write_file(offs, 0, data, 2981);

	offs = create_file("file2");
	write_file(offs, 0, data, 9362);

	dump_files();

	offs = find_file("remove2");
	remove_file(offs);
	offs = find_file("remove1");
	remove_file(offs);
	dump_files();

	offs = find_file("file1");
	compact_files_backward(offs);
	dump_files();

	offs = find_file("file1");
	compact_files_forward(offs);
	dump_files();

	offs = find_file("firstfile");
	remove_file(offs);
	dump_files();

//	offs = find_file("file2");
//	compact_files_backward(offs);
//	dump_files();

	offs = find_file("file2");
	compact_files_forward(offs);
	dump_files();


}
*/
