# Standard Makefile for Nintedo DS projects
# -- Sasq 2005

PREFIX = arm-elf-

CC = $(PREFIX)gcc
LD = $(PREFIX)gcc
AR = $(PREFIX)ar
OBJCOPY = $(PREFIX)objcopy
OBJDUMP = $(PREFIX)objdump

OBJFILES = $(addprefix $(OBJDIR)/,$(OBJS))

CFLAGS = -Wall -O2 -mcpu=arm9tdmi -mtune=arm9tdmi -mthumb-interwork -fomit-frame-pointer -ffast-math -fno-builtin -nostdinc -DARM9
LDFLAGS = -mno-fpu -mthumb-interwork -nostdlib -specs=ds_arm9.specs
ASFLAGS = -marm -x assembler-with-cpp -mthumb-interwork -DARM9



$(OBJDIR) :
	"mkdir" -p $(OBJDIR)

$(TARGET).elf : $(DEPENDS) $(OBJDIR) $(OBJFILES)
	$(LD) $(LDFLAGS) $(OBJFILES) $(LIBS) -lgcc -o $@

$(TARGET).a : $(OBJDIR) $(OBJFILES)
	$(AR) $(ARFLAGS) $@ $(OBJFILES)
	
.PRECIOUS : %.nds %.bin %.fs.bin

%.nds.gba : %.nds
	dsbuild $< $@ 

%.nds : ../arm7/arm7.bin $(TARGET).bin
	ndstool -7 ../arm7/arm7.bin -9 $(TARGET).bin -c $@

$(OBJDIR)/%.o : %.S
	$(CC)  $(ASFLAGS) -c $< -o $@ 
	
$(OBJDIR)/%.o : %.s
	$(CC)  $(ASFLAGS) -c $< -o $@ 
	
$(OBJDIR)/%.o : %.c
	$(CC)  $(CFLAGS) $(DEFINES) $(INCLUDES) -c $< -o $@

%.fs.bin: %.elf
	$(OBJCOPY) -O binary $< data/boot.bin
	..\tools\makefs.exe -b boot.bin -x .svn data $@
	rm data\boot.bin

%.bin: %.elf
	$(OBJCOPY) -O binary $< $@


clean:
	rm -f $(TARGET).elf $(TARGET).nds $(TARGET).bin $(TARGET).nds.gba $(TARGET).a $(OBJFILES)

lclean:
	rm -f $(TARGET).elf $(TARGET).nds $(TARGET).bin $(TARGET).nds.gba $(TARGET).a

superclean: clean
	rm -rf $(OBJDIR) *.bak ~* *.pdb *.ncb *.suo *.vssscc *.opt *.idb
	
rebuild: clean all
	
rb: clean all
	
relink: lclean all

rl: lclean all