
//#include <stdarg.h>

#ifdef _WIN32
#include <io.h>
#include <windows.h>
#endif

//#include <pogo/font.h>

//#include <sys/types.h>
//#include <dirent.h>
//#include <sys/stat.h>

#include <stdio.h>
#include <stdlib.h>

#ifdef POGO_PLATFORM_NDS
#include <nds.h>
#endif

#include <pogo/pogo.h>
#include <pogo/keyboard.h>
#include <pogo/console.h>

void key_init();
int console_findword(int x, int y, char *dest);
void console_markword(int x, int y);

#include "glk.h"

static const char *data_path =":data:fonts:/";

char *find_file(const char *name, const char *path)
{
	struct pogo_stat s;
	static char tmp[256];
	const char *p = path;
	
	if(!p)
		p = data_path;
	
	while(p)
	{
		int len;
		char *pp;
		if((pp = strchr(p, ':')))
			len = pp-p;
		else
			len = strlen(p);
		
		memcpy(tmp, p, len);
		if(len)
		{
			tmp[len] = '/';
			strcpy(&tmp[len+1], name);
		}
		else
			strcpy(tmp, name);
		
		if((p = pp))
			p++;
		
		if(pogo_stat(tmp, &s) == 0)
			return tmp;
	}
	
	return NULL;
}


static int init = 0;

int graphics_init(void **pixels, int *width, int *height, int *stride)
{
	
	display_set_params(320, 240, 0);
	display_init();
	
	display_get_target((Pixel**)pixels, width, height);
	
	*stride = *width;
	
	return 1;
}

void graphics_flush()
{
	display_flush();
}

void graphics_exit()
{
}

int events_getmsec()
{
	return pogo_clock() * 20;
}

static int resize = 0;
static int new_width;
static int new_height;
static int last_time;
static int draw = 0;
static int text_grab = 0;
char outbuffer[80] = "";
static int lastx, lasty, lastb;

// Get a GLK event from the system
int events_get(event_t *glkevt)
{
	int b,x,y;
	
	if(*outbuffer)
	{
		glkevt->type = evtype_CharInput;
		glkevt->val1 = *outbuffer;
		memmove(outbuffer, &outbuffer[1], strlen(outbuffer));
		return 1;
	}

	int c = pogo_getchar();

				
//	key_get_pointer(&b, &x, &y);
	pogo_ioctl(POGO_STDIN, KC_GET_POINTER, &b, &x, &y);
	if(b)
	{
		if(!text_grab && y < 10 && lastb != b)
		{
			pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, NULL, -1, -1, -1, -1, 0);
			text_grab = 1;
				
		}
	}
	else
	{
		if(text_grab)
		{
			char temp[80];
			//char *p = temp;
			if(console_findword(lastx, lasty, temp))
				strcat(outbuffer, temp);
			console_markword(-1, -1);
			
			
#ifdef POGO_PLATFORM_PC
			Pixel *ptr;
			int width, height;
			//display_set(0);
			screen_get_target(&ptr, &width, &height, NULL);			
			pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, ptr, 0, 10, width, height-10, 0);
#else
			if(draw)
				pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, 0x06000000, 0, 10, 256, 182, 0);
			else
				pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, 0x06200000, 0, 114, 256, 78, 0);
#endif

			text_grab = 0;
				
			display_flush();
		}

	}
	
	lastb = b;
	
	if(c == RAWKEY_L && text_grab)
	{
		char temp[80];
		if(console_findword(lastx, lasty, temp))
		{
			strcat(outbuffer, temp);
			strcat(outbuffer, " ");
		}
		c = 0;		
	}
		
	if(text_grab)
	{
		lastx = x;
		lasty = y;
		console_markword(x, y);
		display_flush();
	}
		
	if(c > 0 && c < 0x80)
	{
		glkevt->type = evtype_CharInput;
		glkevt->val1 = keycode_Unknown;
		
		switch(c)
		{
		case RAWKEY_A:
		case RAWKEY_B:
				glkevt->val1 = ' ';
			break;
		case RAWKEY_START:
		case POGOKEY_ENTER: glkevt->val1 = keycode_Return;
			break;
		case POGOKEY_RIGHT: glkevt->val1 = keycode_Right;
			break;
		case POGOKEY_LEFT: glkevt->val1 = keycode_Left;
			break;
		case POGOKEY_UP: glkevt->val1 = keycode_Up;
			break;
		case POGOKEY_DOWN: glkevt->val1 = keycode_Down;
			break;
		case POGOKEY_HOME: glkevt->val1 = keycode_Home;
			break;
		case POGOKEY_END: glkevt->val1 = keycode_End;
			break;
		case POGOKEY_BACKSPACE: glkevt->val1 = keycode_Delete;
			break;
		case POGOKEY_ESCAPE: glkevt->val1 = keycode_Escape;
			break;
		case RAWKEY_X:
		case POGOKEY_PAGEUP: glkevt->val1 = keycode_PageUp;
			break;
		case RAWKEY_Y:
		case POGOKEY_PAGEDOWN: glkevt->val1 = keycode_PageDown;
			break;
		case POGOKEY_TAB: glkevt->val1 = keycode_Tab;
			break;
		case POGOKEY_KILL: 
			glkevt->val1 = keycode_Func12;
		case RAWKEY_L:
		case RAWKEY_R:
			break;
			
#ifdef POGO_PLATFORM_NDS
		case RAWKEY_SELECT:
			if(!draw)
			{
				lcdMainOnBottom();
				pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, 0x06000000, 0, 10, 256, 182, 0);
				pogo_ioctl(POGO_STDIN, KC_ACTIVATE_KEYBOARD, 0);
				draw =1 ;
			}
			else
			{
				lcdMainOnTop();
						//key_set_stroke_target(NULL, -1, -1, -1, -1, 0);
				pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, 0x06200000, 0, 114, 256, 78, 0);
				pogo_ioctl(POGO_STDIN, KC_ACTIVATE_KEYBOARD, 1);
				draw = 0;
			}
			return 0;
#endif
		default:
			glkevt->val1 = c;
			break;
			
		}
		
		return 1;
	}
	
	return 0;
}


void native_init();

void io_init()
{

	pogo_device_init();

#ifdef POGO_PLATFORM_PC
	native_init();
#else
	fatdev_init();
#endif
	key_init();
	console_init();
	pogo_open("/dev/key", 0);
	pogo_open("/dev/console", 0);

	Font *arialfont = POGO_NULL;
	char *fname = find_file("Keyboard.font", NULL);

	if(fname)
		arialfont = font_load(fname);
	if(!arialfont)
		arialfont = font_getdefault();

	font_setcolor(0x000050, 0xafafff, NULL);

#ifdef POGO_PLATFORM_PC

	Pixel *ptr;
	int width, height;
	//display_set(0);
	screen_get_target(&ptr, &width, &height, NULL);
	pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, ptr, 0, 10, width, height-10, 0);

	screen_set(0, 0, width, height, screen_color(0));	
//	screen_set(0, height-192, 256, 10, screen_color(80));

//	pogo_ioctl(POGO_STDIN, KC_SETUP_KEYBOARD, 4, height-192+14, width-4, NULL, arialfont);

	font_setcolor(0x000000, 0xFFE080, NULL);
//	screen_set(0, height-192+114, 256, 78, screen_color(255));	
	display_flush();

	//key_set_stroke_target(ptr, 4, height-192+14, 252, 78, 0);

#else

	screen_set_target((Pixel *)0x06200000, 256, 192);

	screen_set(0, 0, 256, 192, screen_color(0));	
	screen_set(0, 0, 256, 10, screen_color(80));		
	//display_set(0);

	pogo_ioctl(POGO_STDIN, KC_SETUP_KEYBOARD, 4, 14, 256, NULL, arialfont);
	pogo_ioctl(POGO_STDIN, KC_SETUP_GESTURE_TARGET, 0x06200000, 0, 114, 256, 78, 0);

	screen_set_target((Pixel *)0x06200000, 256, 192);
	font_setcolor(0x000000, 0xFFE080, NULL);
	screen_set(0, 114, 256, 78, screen_color(255));	

#endif

	last_time = pogo_clock();

}
