#include "pogo/pogo.h"
#include "pogo/ds_defs.h"
#include "pogo/stroke.h"

extern void getSema(volatile uint32 *sema);
extern void relSema(volatile uint32 *sema);
extern int32 trySema(volatile uint32 *sema);

int console_findword(int x, int y, char *dest);
void console_markword(int x, int y);

static void click2xy(uint16 cx, uint16 cy, int *x, int *y);

static volatile uint32 *reader_sema;

static int mode = 0;

static int in_stroke = 0;

static uchar *lookup;
static uint16 pal[256];
static int idx;

uint32 true2pal(uint16 col)
{
	if(!lookup[col])
	{
		pal[idx] = col;
		lookup[col] = idx++;
	}

	return lookup[col];

}

void draw_to_tile(uint16 *src, uint16 *dst, int width)
{
	int x,y,c0,c1;
	for(y=0; y<8; y++)
	{
		for(x=0;x<4;x++)
		{
			c0 = true2pal(src[0]);
			c1 = true2pal(src[1]);
			*dst = (c1<<8) | c0; 
			//0x7F00;//(src[0] & 0xff); //>>8;// >> 0xff; //(src[0]&0xff) | ((src[1]&0xff)<<8);
			dst++;
			src += 2;
		}
		src += (width-8);
	}
}

void tile16convert(uint16 *src, uint16 *dst)
{
	uint16 *palram =  (uint16*)0x05000400;
	int x,y;
	lookup = malloc(65536);
	memset(lookup, 0, 65536);
	idx = 1;
	
	for(y=0; y<24; y++)
	{
		for(x=0;x<32;x++)
		{
			draw_to_tile(src, dst, 256);
			src += 8;
			dst += 8*4;
		}
		src += (7*256);
	}

	free(lookup);

	for(x=0; x<idx; x++)
		palram[x] = pal[x];

}

#define TO_RGB16(r,g,b) ( ((b << 7) & 0x7C00) | ((g << 2) & 0x03E0) | (r >> 3) )

void draw_square(uint16 *ptr, int size, int width, int flags)
{
	//uint16 col0 = TO_RGB16(0xFF, 0xFF, 0xFF);
	//uint16 col1 = TO_RGB16(0xC0, 0xC0, 0xC0);
	uint16 col0 = 0x0C63;
	uint16 col1 = 0x0C63;

	uint16 col2 = 0x7fff;
	//uint16 col2 = 0x318c;

	int x,y;
	
	if(flags & 1)
		col1 = col2;
	if(flags & 2)
		col0 = col2;

	x = size;
	while(x--)
		*ptr++ = col0;
	
	ptr += (width-size);	
	
	y = size-1;
	
	while(y--)
	{
		*ptr++ = col1;
		x = size-1;
		while(x--)
			*ptr++ = col2;
		ptr += (width-size);
	}
}

void generate_keyboard(Font *font, char *layout, int w, int h, int sqsize, uint16 *vram)
{
	//uint16 *vram = (uint16 *)0x06800000;
	int x,y;
	int flags;
	
	char *lptr = layout;


	font_setcolor(0x000000, 0xFFFFFF);
	
	for(y=0; y<h; y++)
	{
		for(x=0; x<w; x++)
		{
			flags = 0;
			if(x && (*lptr == lptr[-1])) flags = 1;
			if(y && (*lptr == lptr[-w])) flags |= 2;
			
			draw_square(&vram[x*sqsize + y*sqsize*256], sqsize, 256, flags);
			if(!flags)
			{
				int xr = x*sqsize + 6;
				int yr = y*sqsize + 12;
				switch(*lptr)
				{
				case 'H':
					//font_text(font, "Sh", &vram[x*sqsize + 4 + (y*sqsize + 4)*256], 256);
					font_text(font, "Sh", xr, yr);
					break;
				case 'B':
					//font_text(font, "<--", &vram[x*sqsize + 4 + (y*sqsize + 4)*256], 256);
					font_text(font, "<--", xr, yr);
					break;
				case 'E':
					//font_text(font, "Cr", &vram[x*sqsize + 4 + (y*sqsize + 4)*256], 256);
					font_text(font, "Cr", xr, yr);
					break;
				case 'S':
					break;
				default:
					//font_putchar(font, *lptr, &vram[x*sqsize + 4 + (y*sqsize + 4)*256], 256);
					font_putchar(font, *lptr, xr, yr);
					break;
				}
			}
			lptr++;
		}
		
	}

}

//header
#define TOUCH_CAL_X1 (*(volatile uint16*)0x027FFCD8)
#define TOUCH_CAL_Y1 (*(volatile uint16*)0x027FFCDA)
#define TOUCH_CAL_X2 (*(volatile uint16*)0x027FFCDE)
#define TOUCH_CAL_Y2 (*(volatile uint16*)0x027FFCE0)

static PogoDevice vkeydev;

static int oldkeys;

static vuint16 clickbuffer[128];
static vuint16 *inptr = NULL;

static vchar strokebuffer[128];
static vchar *sptr = NULL;

//static volatile uint32 *ipc_sema = (uint32*)0x027FF020;

volatile int vkey_frame_counter = 0;

int strokes = 0;

int sstack[8];

void do_stroke(int x, int y)
{
	int x2,y2,i;

	strokes++;
	if(strokes > 2)
	{
		if(strokes > 5)
		{
			//dprintf("%d %d", sstack[0], sstack[1]);
			stroke_record(sstack[0], sstack[1]);
			click2xy(sstack[0], sstack[1], &x2, &y2);
			//uint16 *scr = &(((uint16 *)(0x06800000))[x2 + y2 * 256]);
			//*scr = *scr ^ 0xffff;
		}

		for(i=0; i<6; i++)
			sstack[i] = sstack[i+2];

		sstack[4] = x;
		sstack[5] = y;
	}


	

}



const char *gestures[][2] =
{
	{ "7[45]2(3)[56]9", "a" }, // A
	{ "7412(5)69", "a" }, // A
	{ "7[45][23]69", "a" }, // A

	{ "(3)2147[58](6)3[56]*", "a" }, // A


	{ "(2)14(7)8[96][58](47)", "b" }, // b
	//{ "14(7)8[96]8(47)", "b" }, // b
	//{ "14(7)865(7)", "b" }, // b

	{ "(3)2(1)4(7)8(9)", "c" }, // c
	//{ "(6)5(47)8(9)63", "d" }, // d
	//{ "878(9)63", "d" }, // d

	{ "(69)[58][47]8(9)63(2)", "d" }, // b


	{ "(47)[85](632)(1)4(7)8(9)", "e" }, //e
//	{ "74(1)23", "f" }, // f
	{ "32(1)47", "f" }, // f
	{ "(3)2(1)4(7)8(9)[65](54)", "g" }, // G
	{ "147(4)5(6)9", "h" }, // h
	{ "14742(3)69", "h" }, // h
	{ "147[85]69", "h" }, // h

	{ "258", "i" }, // I
	{ "(12)36987(4)", "j" }, // J
	//{ "3[25]4[58]9", "k" }, // K
	//{ "324789", "k" }, // K
	//{ "321489", "k" }, // K
	//{ "147[58][36](3)[58]*", "k" }, // K
	{ "3(2)5(8)7", "k" }, // K

	{ "1478(9)", "l" }, // L
	{ "(7)41(4)[58](6)36(9)", "m" }, // M
	{ "74122369", "m" }, // M
	{ "74126(3)6369", "m" }, // M
	{ "741412(3)69", "m" }, // M

	// Common mistake... either A or M
	{ "7412369", "m" }, // M

	{ "(7)41(2)[45](8)96(3)", "n" }, // N
	{ "(7)41(2)[45](6)3", "n" }, // N
	{ "7[45]8(9)63", "n" }, // N

	{ "(53)2(1)4(7)8(9)63(21)", "o" }, // O
	{ "(53)[21](1)4(7)8(9)6(3)2(1)", "o" }, // O

	{ "14(7)8[96][58](47)", "b" }, // b

	{ "74(1)2[36][52](41)", "p" }, // P
	
	{ "96(3)2[14][52](63)", "q" }, // q
	
	{ "14[78]4(1)2(36)", "r" }, // r
	{ "(8)74(1)2[36](6)", "r" }, // r

	{ "(3)2[14][25][69]8(74)", "s" }, // S
	{ "(3)25698(74)", "s" }, // S
	{ "(1)236[89](89)", "t" }, // T
	{ "1478963", "u" }, // U

	{ "1[45]8[56]3", "v" }, // V
	{ "1478[56]3", "v" }, // V
	{ "14(5)8963", "v" }, // V

	{ "(1)47(48)5(86)96(3)", "w" }, // W

	{ "1478963", "w" }, // W


	{ "7(4)5(6)321(4)5(6)9", "x" }, // X

	{ "1[25]6[58]7", "x" }, // X
	{ "123687", "x" }, // X
	{ "126987", "x" }, // X

	{ "1(4)5(6)369(87)", "y" }, // y
	{ "142369", "y" }, // y
	{ "(1)23(2)5(4)78(9)", "z" }, // Z

	{ "654", "\010" },
	{ "456", " " },
	//{ "", "\n" },
	{ "852", "\n" },

	{ "123658", "?" },
	{ "65456", "west\n" },
	{ "45654", "east\n" },
	{ "25852", "south\n" },
	{ "85258", "north\n" },

	{ "15951", "southwest\n" },
	{ "95159", "northeast\n" },
	{ "35753", "southwest\n" },
	{ "75357", "northeast\n" },

	{ "456545(6)", "examine " },
	{ "654565(4)", "look " },
	{ "258525(8)", "open " },
	{ "842484(2)", "close " },
	
/*
	{ "(3)2(14)5(6)98(7)", "stand " }, // S
	{ "(2)1478963(2)", "open " }, // O
	{ "147(8)9(6)5(4)", "attack " }, // b
	{ "7415369", "score\n" }, // M
	{ "1474(12)3", "read " }, // r
	{ "14789", "look\n" },
	{ "74123", "inventory\n" },

	{ "14(7)8(9)63", "get " },
	{ "74(1)2(3)69", "drop " },

	{ "74159", "all " },
	{ "12357", "it " },

	{ "36987", "\n" },

	{ "1235789", "wait\n" },

	{ "32(1)4(7)89", "examine " },
	{ "12(3)6(9)87", "read " },
*/
	//{ "25852", "down\n" },
	//{ "85258", "up\n" },
};

int match(char *seq, char *pattern)
{
	int opt = 0;
	int either = 0;
	int em = 0;
	char *p = pattern;
	char *s = seq;

	while(*p)
	{
		if(*p == '(') opt++;
		else
		if(*p == ')') opt--;
		else
		if(*p == '[') either++;
		else
		if(*p == ']')
		{
			either--;
			if(!em)
				return 0;
			em = 0;
		}
		else
		{
			if(*p == '*')
				return 1;

			if(*s != *p)
			{
				if(!(opt || either))
					return 0;
			}
			else
			{
				if(!*s)
					return 0;

				if(either)
					em++;

				s++;
			}
		}
		p++;
	}

	if(*p || *s)
		return 0;

	return 1;
}

void fixup(char *src, char *dst)
{
	static const char *check[] = {"123","126","147","148","159",
	"247", "258", "259",
	"321", "324", "357", "368", "369",
	"423", "456", "489",
	"621", "654", "687",
	"741", "742", "753", "786", "789",
	"841", "852", "863",
	"987", "984", "951", "962", "963" };

	char *s = src;
	char *d = dst;

	while(*s)
	{
		int i;
		while(s[0] == s[1])
			s++;
		*d++ = *s;
		for(i=0; i<sizeof(check)/4; i++)
		{
			if(s[0] == check[i][0] && s[1] == check[i][2])
			{
				*d++ = check[i][1];
				break;
			}
		}
		s++;
	}
	*d = 0;
}

int end_stroke()
{
	int i;
	char seq[32];
	if(stroke_trans(seq))
	{
		//dprintf("Got stroke %s", seq);

		if(strlen(seq) >= 3)
		{
			char seq2[48];
			fixup(seq, seq2);

			for(i=0; i<sizeof(gestures)/8; i++)
			{
				if(match(seq2, (char *)gestures[i][0]))
				{
					strcpy((char *)sptr, gestures[i][1]);
					sptr += strlen((char *)sptr);
					break;
				}
			}
			if(i == sizeof(gestures)/8)
			{
				font_setcolor(0x00ffffff, 0x0000000);
				strcat(seq2, "        ");
				font_text(font_getdefault(), seq2, 0, 0);
				//strcpy((char *)sptr, seq2);
				//sptr += strlen((char *)sptr);
			}
			return 1;
		}
	}
	return 0;
}




volatile int touchX;
volatile int touchY;

static char *layout;
static int sqsize;
static int width;
static int height;


static void click2xy(uint16 cx, uint16 cy, int *x, int *y)
{
	int TOUCH_WIDTH  = TOUCH_CAL_X2 - TOUCH_CAL_X1;
	int TOUCH_HEIGHT = TOUCH_CAL_Y2 - TOUCH_CAL_Y1;
	int TOUCH_OFFSET_X = ( ((256 - 60) * TOUCH_CAL_X1) / TOUCH_WIDTH  ) - 28;
	int TOUCH_OFFSET_Y = ( ((192 - 60) * TOUCH_CAL_Y1) / TOUCH_HEIGHT ) - 28;

	*x = ( ((256 - 60) * cx) / TOUCH_WIDTH  ) - TOUCH_OFFSET_X;
	*y = ( ((192 - 60) * cy) / TOUCH_HEIGHT ) - TOUCH_OFFSET_Y;
		
	touchX = cx;
	touchY = cy;

	//*y += 20;
	//*x -= 20;

	if((*x)<0) *x=0;
	else if((*x)>255) *x=255;

	if((*y)<0) *y=0;
	else if((*y)>191) *y=191;

}


static char click_convert(uint16 cx, uint16 cy)
{
	int x,y;
	char c;

	click2xy(cx, cy, &x, &y);
	c = layout[(x-2)/sqsize + ((y-8)/sqsize)*width];
	switch(c)
	{
	case 'B':
		return 8;
	case 'S':
		return ' ';
	case 'E':
		return 10;
	default:
		return c;
	}
}




int px, py;


static void vkey_update(void)
{
	uint32 i = *IF;

	if(!inptr)
		return; // Not openened yet

	vkey_frame_counter++;
	//if((vkey_frame_counter & 0xFF) == 0)
	//	debug_print("FRAME");

	if(i & IRQ_TIMER1)
	//if(i & IRQ_VBLANK)
	{
		//getSema(ipc_sema);
		uint16 keys = ~(*KEYS);

		if((IPC16[12] & 0x40) == 0)
		{
			int x,y;

			px = IPC16[2];
			py = IPC16[3];

			//if(!(trySema(reader_sema)))
			//	return;
			//dprintf("Got %x %d %d", keys, IPC16[2], IPC16[3]);

			if(IPC16[12] != oldkeys)
			{
				click2xy(px, py, &x, &y);
				//dprintf("Staring at %d %d", x, y);
				if(y < 170)
					in_stroke = 1;
				else
					in_stroke = 0;
			}
			//in_stroke = 1;

			if(in_stroke)
				do_stroke(px, py);
			else
			{
				click2xy(px, py, &x, &y);
				console_markword(x, y);
			}
			//relSema(reader_sema);
		}
		else
		{

			if(in_stroke)
			{
				end_stroke();
				in_stroke = 0;
			}
			else
			if(IPC16[12] != oldkeys)
			{
				console_markword(-1, -1);
				*inptr++ = px;
				*inptr++ = py;
				
				//dprintf("Got %d %d", IPC16[2], IPC16[3]);
			}
			strokes = 0;
		}
		oldkeys = IPC16[12];
		//relSema(ipc_sema);
	}
}


void vkey_mode(int m)
{
	mode = m;
}

static int vkey_read(int fd, char *dest, int count)
{
	volatile char *cdest2;
	int org = count;
	volatile uint16 *s = clickbuffer;
	volatile char *p = strokebuffer;

	if(!layout || ((inptr == clickbuffer) && (sptr == strokebuffer)))
		return 0;

	//getSema(reader_sema);
	*IME = 0;

	/*if(mode == 0)
	{
		// Copy from buffer to dest, up to count values
		while(s < inptr && count)
		{
			dprintf("Got a click");
			*dest++ = click_convert(s[0], s[1]);
			count--;
			s += 2;
		}

		// Copy unread chars to beginning of buffer
		volatile uint16 *cdest = clickbuffer;
		while(s < inptr)
		{
			*cdest++ = *s++;
		}
		inptr = cdest;

	}
	else*/
	{
		while(s < inptr)
		{
			char theword[128];
			int x, y;
			click2xy(s[0], s[1], &x, &y);
			strcpy(theword, "");
			if(console_findword(x, y, theword))
			{
//				dprintf("Got %s", theword);
				if(strlen(theword))
				{
					strcpy((char *)sptr, theword);
					sptr += strlen(theword);
				}
			}
//			else
//				dprintf("No word at %d %d (%d %d)", x, y, s[0], s[1]);
			s += 2;
		}
		inptr = clickbuffer;
	}

	while(p < sptr && count)
	{
		//dprintf("Got a stroke");
		*dest++ = *p;
		p++;
		count--;
	}
	
	cdest2 = strokebuffer;
	while(p < sptr)
	{
		*cdest2++ = *p++;
	}
	sptr = cdest2;

	//relSema(reader_sema);
	*IME = 1;
	*IF = 0xFFFFFFFF;

	return org - count;
}

/*
int vkey_read2(int fd, char *dest, int count)
{
	int org = count;
	volatile uint16 *s = clickbuffer;

	if(!layout || (inptr == clickbuffer))
		return 0;

	//getSema(reader_sema);
	*IME = 0;

	// Copy from buffer to dest, up to count values
	while(s < inptr && count)
	{
		int x,y;
		click2xy(s[0], s[1], &x, &y);
		*dest++ = x;
		*dest++ = y;
		count -= 2;
	}

	// Copy unread chars to beginning of buffer
	volatile uint16 *cdest = clickbuffer;
	while(s < inptr)
	{
		*cdest++ = *s++;
	}
	inptr = cdest;

	//relSema(reader_sema);
	*IME = 1;
	*IF = 0xFFFFFFFF;

	return org - count;

}
*/

static int vkey_ioctl(int fd, int req, va_list vl)
{
	//char *s;
	//int c;

	switch(req)
	{
	case IO_GETPARAM:
		//(va_arg(vl, int))
		break;
	}
	
	return 0;
}
			


void vkey_generate(Font *font, char *lay, int w, int h, int sq)
{
	uint16 *temp;
	//int i;
	//uint16 *screen = (volatile uint16*)0x06200000;
	uint16 *tiles = (uint16*)0x06204000;
	//uint16 *pal =  (volatile uint16*)0x05000400;

	//Font *f = font_load("/rom/verdana.font");
	//if(!f)
	//Font *f = font_memload(&fixed_font, "fixed");
	
	sqsize = sq;
	width = w;
	height = h;
	layout = lay;

	temp = malloc(256*192*2);
	//temp = 0x02300000;
	//while(1);
	//if(!temp)
	//	return;
	//memset((void*)temp, 0x07, 256*192*2);
	font_settarget(temp, 256, 192, 0);
	generate_keyboard(font, layout, width, height, sqsize, (void*)(temp + 2 + 8*256));
	tile16convert((void*)temp, tiles);
	//memset((void*)tiles, 0x0e, 256*192*2);
	free(temp);
}

int vkey_open(const char *name, int mode)
{
	int i;
	uint16 *tiles = (uint16*)0x06204000;
	uint16 *screen = (uint16*)0x06200000;
	//volatile uint16 *tiles = (volatile uint16*)0x06204000;
	uint16 *palram =  (uint16*)0x05000400;

	//for(i=0; i<32; i++)
	//	palram[i] = i | (i<<5) | (i<<10);

	for(i=0; i<30*30; i++)
		screen[i] = i;// | ((i+1)<<8);

	//for(i=0; i<256*256; i++)
	//	tiles[i] = 0x070e;


	inptr = clickbuffer;

	dprintf("VKEY Opened");

	return 0;

}

#define TIMER_FREQ_256(n) (-(0x2000000>>8)/(n))

void vkey_init(void)
{
	//uint16 l;

	inptr = 0;
	oldkeys = -1;
	reader_sema = malloc(4);
	*reader_sema = 0;
	layout = 0;

	memset(&vkeydev, 0, sizeof(vkeydev));
	vkeydev.open = vkey_open;
	vkeydev.ioctl = vkey_ioctl;
	vkeydev.read = (int (*)(int, void *, int))vkey_read;
	pogo_device_register(&vkeydev, "/dev/vkey", vkey_update, 0);

	*BANK_C_CR = 0x84; // Map 0x06200000 to bank C
	*DISPCNT_SUB = MODE_0_2D | DISPLAY_BG0_ACTIVE;
	*BG0CNT_SUB = 0x80 | 1<<2; // SCRBASE AT 0x0620000, CHARBASE AT 0x06204000

	*TM1CNT_L = TIMER_FREQ_256(100);
	*TM1CNT_H = 0xC2;
	*IE = IRQ_TIMER1 | IRQ_VBLANK;

	stroke_init();

}
