#include <nds.h>

#include <pogo/device.h>

#undef NDS
#define NDS
#include "rein/gba_nds_fat.h"
#undef PACKED

#include <stdio.h>
#include <string.h>

PogoDevice fatdev;

static FAT_FILE *open_files[32];
static int init = 0;

char tmpname[256];

int fatdev_open(const char *name, int flags)
{
	FAT_FILE *f;

	if(!init)
	{
		memset(open_files, 0, sizeof(FAT_FILE*) * 32);
		init = 1;
	}

	//mod_volume(0);

	FAT_chdir("/");

	while(*name == '/') name++;
	sprintf(tmpname, "/%s", name);

	char *fg;

	if(flags == PGO_RDONLY)
		fg = "r";
	else
	if(flags == PGO_RDWR)
		fg = "r+";
	else
	if(flags == PGO_RDWR|PGO_CREAT)
		fg = "w+";
	else
	if(flags & PGO_WRONLY)
		fg = "w";


	f = FAT_fopen(tmpname, fg);

	if(f)
	{
		int i = 0;
		while(open_files[i]) i++;
		open_files[i] = f;
		return i;
	}
	//mod_volume(127);
	return -1;
}

int fatdev_read(int fd, void *dest, int size)
{
	FAT_FILE *f = open_files[fd];
	return FAT_fread(dest, 1, size, f);
}

int fatdev_write(int fd, const void *dest, int size)
{
	FAT_FILE *f = open_files[fd];
	return FAT_fwrite(dest, 1, size, f);
}

void fatdev_close(int fd)
{
	FAT_FILE *f = open_files[fd];
	FAT_fclose(f);
	open_files[fd] = NULL;
	//mod_volume(127);
}

int fatdev_lseek(int fd, int offset, int origin)
{
	FAT_FILE *f = open_files[fd];

	FAT_fseek(f, offset, origin);
	return FAT_ftell(f);
}

int fatdev_tell(int fd)
{
	FAT_FILE *f = open_files[fd];
	return FAT_ftell(f);
}

int fatdev_ioctl(int fd, int request, ...)
{
	return -1;
}

int fatdev_remove(const char *name)
{
	FAT_chdir("/");
	while(*name == '/') name++;
	sprintf(tmpname, "/%s", name);
	return FAT_remove(tmpname);
}

int fatdev_mkdir(const char *name)
{
	FAT_chdir("/");
	while(*name == '/') name++;
	sprintf(tmpname, "/%s", name);
	return FAT_mkdir(tmpname);
}

static int count = 0;

int fatdev_dopen(const char *name)
{
	while(*name == '/') name++;
	sprintf(tmpname, "/%s", name);
	count = 0;
	if(FAT_chdir(tmpname))
		return 0;
	return -1;
}

int fatdev_dread(int fd, void *d, int len)
{
	char dest[256];
	char *ptr;
	FILE_TYPE rc;

	if(!count)
		rc = FAT_FindFirstFileLFN(dest);
	else
		rc = FAT_FindNextFileLFN(dest);

	ptr = dest;

	if(!*dest)
	{
		strcpy(dest, "(null)");
	}
	else
	{
		while(*ptr)
		{
			if(*ptr < 0x20 || *ptr > 0x7e)
				*ptr = 'X';
			ptr++;
		}
	}

	count++;

	strcpy((char *)d, dest);

	if(rc == FT_NONE)
		return 0;
	//else
	//	FAT_GetLongFilename(dest);

	return strlen(dest);
}

void fatdev_dclose(int fd)
{
}

int fatdev_stat(const char *name, struct pogo_stat *buffer)
{
	u32 len;
	FAT_chdir("/");

	while(*name == '/') name++;
	sprintf(tmpname, "/%s", name);

	if(FAT_FileExists(tmpname))
	{
		buffer->pst_size = FAT_GetFileSize();
		buffer->pst_mode = 0;
		if(FAT_GetFileAttributes() & ATTRIB_DIR)
			buffer->pst_mode = PGS_IFDIR;
		return 0;
	}
	return -1;
}

int fatdev_init()
{
	memset(&fatdev, 0, sizeof(PogoDevice));
	fatdev.open = fatdev_open;
	fatdev.read = fatdev_read;
	fatdev.write = fatdev_write;
	fatdev.close = fatdev_close;
	fatdev.lseek = fatdev_lseek;
	fatdev.stat = fatdev_stat;
	fatdev.dopen = fatdev_dopen;
	fatdev.dread = fatdev_dread;
	fatdev.dclose = fatdev_dclose;
	fatdev.remove = fatdev_remove;
	fatdev.mkdir = fatdev_mkdir;

	int rc = FAT_InitFiles();

	if(rc)
		pogo_device_register(&fatdev, "/fat", NULL, -1);

	return rc;
}
