
#include <nds.h>
#include <pogo/pogo.h>
#include <stdio.h>
#include <gbfs.h>

PogoDevice gbfsdev;

GBFS_FILE *gbfs_root = NULL;

typedef struct
{
	void *ptr;
	int pos;
	int len;
} gbfile;

static gbfile *open_files[32];
static int open_count = 0;

void cprintf2(const char *, ...);

int gbfsdev_open(const char *name, int flags)
{
	if(gbfs_root)
	{
		u32 len;

		while(*name == '/') name++;

		void *ptr =  (void *)gbfs_get_obj(gbfs_root, name, &len);

		//cprintf2("Opened %s (%p) %p %d %d\n", name, ptr, gbfs_root, gbfs_root->magic[0], gbfs_root->dir_nmemb);
		//while(pogo_getchar() != RAWKEY_B);

		if(ptr)
		{
			gbfile *gb = malloc(sizeof(gbfile));
			gb->len = len;
			gb->ptr = ptr;
			gb->pos = 0;
			open_files[open_count++] = gb;
			return open_count-1;
		}
	}

	return -1;
	//return (int)open(name, 0);
}

int gbfsdev_read(int fd, void *dest, int size)
{
	int left;
	gbfile *gbf = open_files[fd];
	left = gbf->len - gbf->pos;

	//dprintf("read %d (%d left)", size, left);

	if(!left)
		return -1;

	if(size > left)
		size = left;
	memcpy(dest, &((char *)gbf->ptr)[gbf->pos], size);
	gbf->pos += size;

	//dprintf("%02x %02x", ((char *)dest)[0], ((char *)dest)[1]);

	return size;
	//return read(fd, dest, size);
	//return fread(dest, 1, size, (FILE *)fd);
}

int gbfsdev_write(int fd, const void *dest, int size)
{
	//return write(fd, dest, size);
	return -1;
}

void gbfsdev_close(int fd)
{
	free(open_files[fd]);
	//fclose((FILE *)fd);
	//close(fd);
}

int gbfsdev_lseek(int fd, int offset, int origin)
{
	//char *ONAMES[3] = { "SET", "CUR", "END" };
	gbfile *gbf = open_files[fd];

	//dprintf("Seek %d %s", offset, ONAMES[origin]);
	switch(origin)
	{
	case PG_SEEK_SET:
		gbf->pos = offset;
		break;
	case PG_SEEK_CUR:
		gbf->pos += offset;
		break;
	case PG_SEEK_END:
		gbf->pos = gbf->len - offset;
		break;
	}

	if(gbf->pos < 0)
		gbf->pos = 0;
	if(gbf->pos > gbf->len)
		gbf->pos = gbf->len;

	return gbf->pos;

	//return fseek((FILE *)fd, offset, SEEK_SET);
	//return lseek(fd, offset, origin);
}

int gbfsdev_tell(int fd)
{
	gbfile *gbf = open_files[fd];
	return gbf->pos;
	//return -1;
}

int gbfsdev_ioctl(int fd, int request, ...)
{
	return -1;
}

int gbfsdev_remove(const char *name)
{
	return -1;
}

static int current_i;

int gbfsdev_dopen(const char *name)
{
	current_i = 0;

	return 0;
}

int gbfsdev_dread(int fd, void *dest, int len)
{
	if(current_i >= gbfs_count_objs(gbfs_root))
		return -1;
	gbfs_get_nth_obj(gbfs_root, current_i, dest, NULL);
	current_i++;
	return strlen(dest);
}

void gbfsdev_dclose(int fd)
{
	//return 0;
}

int gbfsdev_stat(const char *name, struct pogo_stat *buffer)
{
	u32 len;
	while(*name == '/') name++;
	void *ptr =  (void *)gbfs_get_obj(gbfs_root, name, &len);
	if(ptr)
	{
		buffer->pst_size = len;
		buffer->pst_mode = 0;
		return 0;
	}

	return -1;

}

extern u32 _end;

void *gbfsdev_init_args(char *name, int unit)
{
	if(unit)
		return NULL;
	
	memset(&gbfsdev, 0, sizeof(PogoDevice));
	gbfsdev.open = gbfsdev_open;
	gbfsdev.read = gbfsdev_read;
	gbfsdev.write = gbfsdev_write;
	gbfsdev.close = gbfsdev_close;
	gbfsdev.lseek = gbfsdev_lseek;
	gbfsdev.stat = gbfsdev_stat;
	gbfsdev.dopen = gbfsdev_dopen;
	gbfsdev.dread = gbfsdev_dread;
	gbfsdev.dclose = gbfsdev_dclose;

	pogo_device_register(&gbfsdev, name, NULL, -1);

	u32 *start = (u32 *)((u32)(&_end) && 0xFFFFFF00);

	gbfs_root = find_first_gbfs_file(start);//&_end);
//	gbfs_root = (const GBFS_FILE *)gbfs_space;

//	cprintf2("INIT %p %d %d\n", gbfs_root, gbfs_root->magic[0], gbfs_root->dir_nmemb);

	return gbfs_root;
}

void *gbfsdev_init()
{
	return gbfsdev_init_args("/rom", 0);
}
