
#include "pogo/platform.h"

#ifdef POGO_PLATFORM_NDS

#include <nds.h>
#define IS_NDS

#endif

#include "pogo/pogo.h"
#include "pogo/keyboard.h"

#ifdef USE_NDS
#define LOCK() (REG_IME = IME_DISABLE)
#define UNLOCK() (REG_IME = IME_ENABLE)
#else
#define LOCK()
#define UNLOCK()
#endif


extern vchar strokebuffer[128];
extern vchar *sptr;
extern int strokes;
void end_stroke();
void do_stroke(int x, int y);

void *sa_ptr = POGO_NULL;
int sa_flags = 0;
int sa_x0 = -1;
int sa_x1 = -1;
int sa_y0 = -1;
int sa_y1 = -1;


int stroke_ended = 0;

typedef struct
{
	int x0,x1,y0,y1;
	int width;
	int active;
	uchar *letters;
} Area;

Area **areas;
int acount = 0;

const char *default_line[] = {
	//" x3 f1 12 22 32 42 f1 52 62 72 82",
	//"",
	"1!22@23#24$25%26^27&28*29(20)2-_2 b3",
	" t3qQ2wW2eE2rR2tT2yY2uU2iI2oO2pP2=+2",
	"aA2sS2dD2fF2gG2hH2jJ2kK2lL2;:2'\"2 e3",
	" s3zZ2xX2cC2vV2bB2nN2mM2,<2.>2[{2]}2",
	" c3 a3  0\\|2/?2`~2 x3",
	POGO_NULL
			
};// /?2,`~2" };

const char **line;

	
static int keyxpos;
static int keyypos;
static Font *keyfont;
static int keywidth;
void *keyb_pixels;


static uint16 *click_map0 = POGO_NULL;
static uint16 *click_map1 = POGO_NULL;
static int click_map_size = 0;
static int click_map_width;
static int keyshift = 0;
static int click_map_active = 0;

enum {KEY_SHIFT = 0xfff0, KEY_CTRL, KEY_ALT, KEY_CAPS, KEY_UNSHIFT};

static PogoDevice keydev;

static int oldkeys;
static int old_buttons;
volatile unsigned char *inptr = POGO_NULL;
static volatile unsigned char inbuf[128];

volatile uint16 *clickptr = POGO_NULL;
static volatile uint16 clickbuf[256];


volatile int key_frame_counter = 0;

static int touch_count = 0;

void render_keyboard(int shift);


uint16 *saveptr = POGO_NULL;

void screen_save()
{
	int width, height;
	uint16 *ptr;
	screen_get_target((void *)&ptr, &width, &height, POGO_NULL);

	if(!saveptr)
		saveptr = malloc(width * height * 2);
	memcpy(saveptr, sa_ptr, width*height*2);
}

void screen_load()
{
	int width, height;
	uint16 *ptr;
	screen_get_target((void *)&ptr, &width, &height, POGO_NULL);

	if(!saveptr)
		return;
	memcpy(sa_ptr, saveptr, width*height*2);
}


void convert_xy(int *x, int *y)
{
	int w,h,r;
	int xx = *x;
	int yy = *y;
	
	screen_get_target(POGO_NULL, &w, &h, &r);
	
	if(r == ROT_TOPRIGHT)
	{
		*x = yy;
		*y = h - xx;
		
	}	
	else
		if(r == ROT_BOTTOMRIGHT)
	{
		*x = w - xx;
		*y = h - yy;
		
	}
	else
		if(r == ROT_BOTTOMLEFT)
	{
		*x = w - yy;
		*y = xx;
		
	}
	else
	{
		*x = xx;
		*y = yy;
	}
}

#ifdef IS_NDS

static int lastx = -1;
static int lasty;

volatile int touchx = 0;
volatile int touchy = 0;
volatile int buttons = 0;

int key_get_pointer(int *b, int *x, int *y)
{	
	if(b)
		*b = buttons & KEY_TOUCH;
	if(x)
		*x = touchx;
	if(y)
		*y = touchy;
	
	return 0;
				
}

static void key_update(void)
{
	uint32 i = REG_IF;

	if(!inptr)
		return; // Not openened yet

	key_frame_counter++;

	if(i & IRQ_VBLANK)
	{
		//if(!trySema(reader_sema))
		//	return ;
		int keys = (~(REG_KEYINPUT) & 0x3ff) | (((IPC->buttons)&0x3)<<10);

		if(oldkeys == -1)
			oldkeys = keys;
		else
		if(keys != oldkeys)
		{
			int diff = keys ^ oldkeys;
			oldkeys = keys;

			if(diff & 0x10)
				*inptr++ =  RAWKEY_RIGHT | ((keys & 0x10) ? 0 : 0x80);
			if(diff & 0x20)
				*inptr++ = RAWKEY_LEFT | ((keys & 0x20) ? 0 : 0x80);
			if(diff & 0x80)
				*inptr++ = RAWKEY_DOWN | ((keys & 0x80) ? 0 : 0x80);
			if(diff & 0x40)
				*inptr++ = RAWKEY_UP | ((keys & 0x40) ? 0 : 0x80);
			if(diff & 0x1)
				*inptr++ = RAWKEY_A | ((keys & 0x1) ? 0 : 0x80);
			if(diff & 0x2)
				*inptr++ = RAWKEY_B | ((keys & 0x2) ? 0 : 0x80);
			if(diff & 0x4)
				*inptr++ = RAWKEY_SELECT | ((keys & 0x4) ? 0 : 0x80);
			if(diff & 0x8)
				*inptr++ = RAWKEY_START | ((keys & 0x8) ? 0 : 0x80);
			if(diff & 0x200)
				*inptr++ = RAWKEY_L | ((keys & 0x200) ? 0 : 0x80);
			if(diff & 0x100)
				*inptr++ = RAWKEY_R | ((keys & 0x100) ? 0 : 0x80);
			if(diff & 0x400)
				*inptr++ = RAWKEY_X | ((keys & 0x400) ? 0 : 0x80);
			if(diff & 0x800)
				*inptr++ = RAWKEY_Y | ((keys & 0x800) ? 0 : 0x80);
		}

		scanKeys();
		//int buttons = IPC->buttons;
		buttons = keysHeld();
		touchx = IPC->touchXpx;
		touchy = IPC->touchYpx;
		
		convert_xy(&touchx, &touchy);

		
		if(buttons & KEY_TOUCH)
			touch_count++;
		else
		{
			touch_count = 0;
			if(strokes)
			{
				stroke_ended = 1;
				if(sa_ptr)
				{
					screen_load();
					//display_flush();
				}
				strokes = 0;

			}
			lastx = -1;
		}
		
		if(touchx >= sa_x0 && touchx < sa_x1 && touchy >= sa_y0 && touchy < sa_y1 && (!stroke_ended) && (touch_count >= 2))
		{
			if(sa_ptr && !strokes)
				screen_save();
					
			do_stroke(touchx,  touchy);
					
//			if(lastx >= 0)
	//			draw_line(lastx, lasty, touchx, touchy);
			lastx = touchx;
			lasty = touchy;
			//display_flush();
		}
		else
		if(touch_count == 2)
		{
			int i;

			*clickptr++ = touchx;
			*clickptr++ = touchy;
			//if(i == acount)
				//*inptr++ = acount+1;
		}
	
	}
}

#endif


int key_peek()
{
	return (inptr - inbuf) + (clickptr-clickbuf);
}


static int key_read(int fd, char *dest, int count)
{
	int org = count;
	volatile unsigned char *d, *s;

#ifdef USE_SDL
	sdlevents_get();
#endif

	if(stroke_ended)
	{
		LOCK();
		end_stroke();
		stroke_ended = 0;
		strokes = 0;
		UNLOCK();
	}
		
	if(sptr != strokebuffer)
	{
		LOCK();
		vchar *sb = strokebuffer;
		
		while(count && sptr > sb)
		{
			*dest++ = *sb++;
			count--;
		}
		
		vchar *sb2 = strokebuffer;
		while(sptr > sb)
		{
			*sb2++ = *sb++;
		}
		sptr = sb2;//strokebuffer;

		UNLOCK();
		return org-count;

	}

	if(inptr == inbuf && clickptr == clickbuf)
		return 0;
		

	//dprintf("go\n");

#ifdef USE_NDS
	//getSema(reader_sema);
	REG_IME = IME_DISABLE;
#endif

	volatile uint16 *cp = clickbuf;

	while(clickptr > cp)
	{
		int i;
		int x = *cp++;
		int y = *cp++;
		int found = 0;

		for(i=0; i<acount; i++)
		{
			Area *area = areas[i];
			if(area->active && x >= area->x0 && x <= area->x1 && y >= area->y0 && y <= area->y1)
			{
				int w = (area->x1 - area->x0)/area->width;
				int o = (x-area->x0)/area->width + ((y-area->y0)/area->width)*w;
				*inptr++ = area->letters[o];
				found++;
			}
		}
		
		if(!found && click_map0 && click_map_active && y >= keyypos)
		{
			int offs = ((x - keyxpos) * 2 / keywidth) + ((y-keyypos)/keywidth)*click_map_width;
			
			if(offs >= 0 && offs < click_map_size)
			{
				uint16 c;
				if(keyshift)
					c = click_map1[offs];
				else
					c = click_map0[offs];
				
				switch(c)
				{
				case KEY_SHIFT:
					keyshift = 1;
					render_keyboard(keyshift);
					display_flush();
					break;
				case KEY_CAPS:
					keyshift = 9999;
					render_keyboard(keyshift);
					display_flush();
					break;
				case KEY_UNSHIFT:
					keyshift = 0;
					render_keyboard(keyshift);
					display_flush();
					break;
				default:
					*inptr++ = c;
					if(keyshift)
					{
						keyshift--;
						if(!keyshift)
						{
							render_keyboard(0);
							display_flush();
						}
					}
				}
				
			}
		}
	}

	clickptr = clickbuf;
	
	s = inbuf;

	/* Copy from buffer to dest, up to count values */
	while(s < inptr && count)
	{
		/*if(*s == RAWKEY_L && click_map_active && keyshift == 0)
		{
			keyshift = 1;
			render_keyboard(1);
			display_flush();			
		}
		else
		if(*s == (RAWKEY_L|0x80) && click_map_active && keyshift == 1)
		{
			keyshift = 0;
			render_keyboard(0);
			display_flush();			
	}*/
		
		*dest++ = *s++;
		count--;
	}

	/* Copy unread chars to beginning of buffer */
	d = inbuf;
	while(s < inptr)
	{
		*d++ = *s++;
	}
	inptr = d;

#ifdef USE_NDS
	//relSema(reader_sema);
	REG_IME = IME_ENABLE;
#endif


	return org-count;
}

int key_open(const char *name, int mode)
{
	int i;

	if(strlen(name) > 0)
		return -1;

	inptr = inbuf;
	clickptr = clickbuf;

	//dprintf("KEY Opened");

	return 0;

}


int add_clickarea(int x, int y, int w, int h, int charw, uchar *values)
{
	Area *a = malloc(sizeof(Area));

	a->x0 = x;
	a->y0 = y;
	a->x1 = x+w;
	a->y1 = y+h;	
	a->letters = values;
	a->width = charw;
	a->active = 1;

	areas[acount++] = a;

	return acount-1;
}

void key_activate_clickarea(int area, int on)
{
	areas[area]->active = on;
}


//#define TIMER_FREQ_256(n) (-(0x2000000>>8)/(n))

extern void InterruptProcess(void);


void draw_button(int x, int y, int w, int h)
{
	screen_set(x+1, y+1, w-2, h-2, screen_color(0));
	
	if(1)
	{
		screen_set(x, y+1, 1, h-2, screen_color(255));
		screen_set(x+1, y, w-2, 1, screen_color(255));
	
		screen_set(x+1, y+h-1, w-2, 1, screen_color(255));
		screen_set(x+w-1, y+1, 1, h-2, screen_color(255));
	}
	else
	{
		screen_set(x, y, 1, h, screen_color(64));
		screen_set(x, y, w, 1, screen_color(64));
	
		screen_set(x+1, y+h-1, w-1, 1, screen_color(128));
		screen_set(x+w-1, y+1, 1, h-1, screen_color(128));
	}
}

void key_activate_keyboard(int on)
{
	click_map_active = on;
}

	
void render_keyboard(int shift)
{
	int col_change = 1;
	int i,j, w, width, height;
	void *ptr;
	char tmp[3] = "Fx";
	int x = keyxpos;
	int y = keyypos;
	screen_get_target(&ptr, &width, &height, POGO_NULL);
	
	//display_set(0);
	
	font_setcolor(0x000000, 0xFFFFFF, keyfont);
	//#ifdef IS_NDS
	screen_set_target(keyb_pixels, 0, 0);	
	//#endif
	
	//int count = linesizeof(line)/sizeof(char*);
	
	for(j=0; line[j]; j++)
	{					
		int o = x;
		
		for(i=0; i<strlen(line[j])/3; i++)
		{
			
			char s = 0;
			if(line[j][i*3] == ' ')
				s = line[j][i*3+1];
			w = (line[j][i*3+2] - 0x30) * keywidth / 2;
			if(!w) w = keywidth * 5;

			if(s != 'f')
			{
				int p = line[j][i*3+(shift ? 1 : 0)];
				char *str = POGO_NULL;
				
				if(s == 's' && keyshift == 1)
				{
					font_setcolor(0x000000, 0xFFFF80, keyfont);
					col_change++;
				}
				else
				if(s == 'C' && keyshift > 1000)
				{
					font_setcolor(0x000000, 0xFFFF80, keyfont);
					col_change++;
				}
				else
					if(col_change)
				{
					font_setcolor(0x000000, 0xFFFFFF, keyfont);
					col_change = 0;
				}
				
				draw_button(o, y, w-2, keywidth-2);
				
				switch(s)
				{
					case 'e':						
						p = 0x21b5;
						if(font_charwidth(keyfont, p) <= 0)
							str = "Ret";
						break;
					case 'b':
						p = 0x2190;
						if(font_charwidth(keyfont, p) <= 0)
							str = "<-";
						break;
					case 's':
						p = 0x21e7;
						if(font_charwidth(keyfont, p) <= 0)
							str = "Sh";
						break;
					case 't':
						p = 0x21e5;
						if(font_charwidth(keyfont, p) <= 0)
							str = "Tab";
						break;
					case 'c':
						str = "Ctl";
						break;
					case 'a':
						str = "Alt";
						break;
					case 'x':
						str = "Esc";
						break;
					case 'C':
						str = "Cps";
						break;
					case 'U':
						str = "pUp";
						break;
					case 'D':
						str = "pDn";
						break;
					case 'h':
						str = "Hom";
						break;
					case 'n':
						str = "End";
						break;
				}
				
				if(s >= '0' && s <= '9')
				{
					tmp[1] = s;
					str = tmp;
				}
					
				if(str)
				{
					int fw = font_text(keyfont, str, -1, -1);
					font_text(keyfont, str, o + (w - fw - 1)/2, y + (keywidth - keyfont->height)/2);
				}
				else
				{
					int fw = font_putchar(keyfont, p, -1, -1);
					font_putchar(keyfont, p, o + (w - fw - 1)/2, y + (keywidth - keyfont->height)/2);
				}
			}
			
			o += w;
		}
		
		y += keywidth;
		
	}

	screen_set_target(ptr, 0, 0);
	
}

void key_set_stroke_target(void *ptr, int x, int y, int w, int h, int flags)
{
	sa_ptr = ptr;
	sa_x0 = x;
	sa_y0 = y;
	sa_x1 = x+w;
	sa_y1 = y+h;
	sa_flags = flags;
}

int key_setup_keyboard(int x, int y, int w, const char **def, Font *font)
{
	int width, addx;
			
	int i,j;
	
	if(!def)
	{
		line = default_line;
		//count = sizeof(default_line)/sizeof(char*);
	}
	else
		line = def;
	
	screen_get_target(&keyb_pixels, NULL, NULL, NULL);
	
	
	if(!font)
		font = font_getdefault();
				
	//keywidth = 20;
	keyxpos = x;
	keyypos = y;
	keyfont = font;
	
	int len = 0;
	
	click_map_active = 1;
	
	for(j=0; line[j]; j++)
	{
		int x = 0;
		for(i=0; i<strlen(line[j])/3; i++)
		{
			int addx = (line[j][i*3+2] - 0x30);
			if(!addx) addx = 10;			 
			x += addx;
		}
		if(x > len)
			len = x;
	}
	
	keywidth = (w * 2 / len);
	
	width = len*2;
	
	keyxpos += (w - keywidth * len / 2) / 2;
	
	click_map0 = malloc(width*7*2);
	click_map1 = malloc(width*7*2);
	memset(click_map0, 0, width*7*2);
	memset(click_map1, 0, width*7*2);
	
	click_map_size = width*7;
	click_map_width = width;
			
	y = 0;
	
	for(j=0; line[j]; j++)
	{					
		x = 0;
		int len = strlen(line[j]);
		
		for(i=0; i<len/3; i++)
		{
			char s = 0;
			if(line[j][i*3] == ' ')
				s = line[j][i*3+1];
			addx = (line[j][i*3+2] - 0x30);
			if(!addx) addx = 10;

			int p = line[j][i*3];
			int q = line[j][i*3+1];
			
			if(s == 'f')
			{
				p = q = 0;
			}

							
			switch(s)
			{
				case 'e':
					p = 10;
					q = 13;
					break;
				case 'b':
					p = 8;
					q = 25;
					break;
				case 's':
					p = KEY_SHIFT;
					q = KEY_UNSHIFT;
					break;
				case 'C':
					p = KEY_CAPS;
					q = KEY_UNSHIFT;
					break;
				case 't':
					p = 9;
					break;
				case 'c':
					p = KEY_CTRL;
					break;
				case 'a':
					p = KEY_ALT;
					break;
				case 'x':
					p = 27;
					break;
			}
			
			if(s >= '0' && s <= '9')
			{
				//tmp[1] = s;
				//str = tmp;
			}
								
			x += addx;
			
			while(addx--)
			{
				click_map0[y*width+x-addx-1] = p;
				click_map1[y*width+x-addx-1] = q;
				
			}
		}
		
		y += 1;
		
	}

	render_keyboard(0);
	
	return y * keywidth;
}


static int key_ioctl(int fd, int req, va_list vl)
{
	switch(req)
	{
		case KC_ACTIVATE_KEYBOARD:
			key_activate_keyboard(va_arg(vl, int));
			return;
		case KC_ACTIVATE_CLICKAREA:
			key_activate_clickarea(va_arg(vl, int), va_arg(vl, int));
			return;
		case KC_SETUP_KEYBOARD:
		{
			int x = va_arg(vl, int);
			int y = va_arg(vl, int);
			int w = va_arg(vl, int);
			const char **def = va_arg(vl, const char**);
			Font *font = va_arg(vl, Font*);
			return key_setup_keyboard(x, y, w, def, font);
		}
		case KC_SET_SHIFT:
			keyshift = va_arg(vl, int);
			return;
		case KC_GET_SHIFT:
			return keyshift;
		case KC_REFRESH_KEYBOARD:
			render_keyboard(keyshift);
			return;
		case KC_ADD_CLICKAREA:
		{
			int x = va_arg(vl, int);
			int y = va_arg(vl, int);
			int w = va_arg(vl, int);
			int h = va_arg(vl, int);
			int charw = va_arg(vl, int);
			uchar *values = va_arg(vl, char*);
			return add_clickarea(x, y, w, h, charw, values);
		}
		case KC_GET_POINTER:
		{
			int *b = va_arg(vl, int*);
			int *x = va_arg(vl, int*);
			int *y = va_arg(vl, int*);
			return key_get_pointer(b, x, y);
		}			
		case KC_SETUP_GESTURE_TARGET:
		{
			char *ptr = va_arg(vl, char*);
			int x = va_arg(vl, int);
			int y = va_arg(vl, int);
			int w = va_arg(vl, int);
			int h = va_arg(vl, int);
			int flags = va_arg(vl, int);
			key_set_stroke_target(ptr, x, y, w, h, flags);
			return;
		}
	}
	
	return 0;
}



void sdlkey_init();

void key_init(void)
{
	//uint16 l;
	
	sptr = strokebuffer;

	inptr = 0;
	oldkeys = -1;
	//reader_sema = malloc(4);
	//*reader_sema = 0;

	areas = malloc(sizeof(Area *) * 32);

	memset(&keydev, 0, sizeof(keydev));
	keydev.open = key_open;
	keydev.ioctl = key_ioctl;
	keydev.read = (int (*)(int, void *, int))key_read;
#ifdef IS_NDS
	pogo_device_register(&keydev, "/dev/key", key_update, 0);
#else
	pogo_device_register(&keydev, "/dev/key", POGO_NULL, 0);
	sdlkey_init();
#endif

//TIMER1_CR
	//TIMER1_DATA = TIMER_FREQ_256(100);
	//TIMER1_CR = 0xC2;
	//*IE = IRQ_TIMER1 | IRQ_VBLANK;

	//irqSet(IRQ_TIMER1, InterruptProcess);
	//irqEnable(IRQ_VBLANK | IRQ_TIMER1);


}
