/*
 * PSP Filesystem Device (pspfile.c)
 * Part of libpogo, a c-library replacent for consoles
 * Programmed by Jonas Minnberg (Sasq)
 *
 * DESCRIPTION
 **/

#include "core.h"
#include "device.h"

#include "psp.h"

static Device pspdev;

int sceIoOpen(const char* file, int mode, int unknown); 
void sceIoClose(int fd); 
int sceIoRead(int fd, void *data, int size); 
int sceIoWrite(int fd, void *data, int size); 
long long sceIoLseek(int fd, long long offset, int whence);
int sceIoRemove(const char *file); 
int sceIoMkdir(const char *dir, int mode); 
int sceIoRmdir(const char *dir); 
int sceIoRename(const char *oldname, const char *newname); 
//int sceIoDevctl(const char *name int cmd, void *arg, size_t arglen, void *buf, 

int psp_open(const char *name, int flags)
{
	char tmp[128];
	strcpy(tmp, "ms0:/");
	strcpy(&tmp[5], name);
	//sprintf(tmp, "ms0:/%s", name);
	dprintf("Opening %s (%d)\n", tmp, flags);
	
	if(flags & O_WRONLY)
		return sceIoOpen(tmp, flags, 0777);
	else
		return sceIoOpen(tmp, flags, 0);
	//file = sceIoOpen(savePath,O_CREAT | O_TRUNC | O_RDWR, 0777);
}

int psp_read(int fd, void *dest, int size)
{
	
	return sceIoRead(fd, dest, size);
}

int psp_write(int fd, const void *dest, int size)
{
	return sceIoWrite(fd, (void *)dest, size);
}

void psp_close(int fd)
{
	sceIoClose(fd);
}

int psp_lseek(int fd, int offset, int origin)
{
	return sceIoLseek(fd, offset, origin);
}

int psp_ioctl(int fd, int request, ...)
{
	return -1;
}

int psp_remove(const char *name)
{
	char tmp[128];
	strcpy(tmp, "ms0:/");
	strcpy(&tmp[5], name);
	return sceIoRemove(tmp);
}

int psp_stat(const char *name, struct stat *buffer)
{
	SceIoStat statbuf;
	char tmp[128];
	strcpy(tmp, "ms0:/");
	strcpy(&tmp[5], name);
	memset(&statbuf, 0, sizeof(SceIoStat));

	dprintf("psp stat %s\n", tmp);

	if(sceIoGetstat(tmp, &statbuf) >= 0)
	{
		buffer->st_size = statbuf.st_size;
		buffer->st_mode = statbuf.st_mode;
		return 0;
	}

	return -1;
}

int psp_dopen(const char *name)
{
	int rc;
	char tmp[128];
	strcpy(tmp, "ms0:/");
	strcpy(&tmp[5], name);
	rc = sceIoDopen(tmp);
	dprintf("Opening %s -> %d\n", name, rc);
	return rc;
}


int psp_dread(int fd, void *dest, int size)
{
	int rc;
	SceIoDirent scede;
	
	memset(&scede, 0, sizeof(scede));
	
	dprintf("Reading %d -> %p\n", fd, dest);
	
	if((rc = sceIoDread(fd, &scede)) > 0)
	{
		strcpy(dest, scede.d_name);
		dprintf("%s\n", dest);
		return strlen(dest);
	}
	return rc;
}

void psp_dclose(int fd)
{
	sceIoDclose(fd);
}



int pspfile_init()
{
	memset(&pspdev, 0, sizeof(pspdev));
	pspdev.open = psp_open;
	pspdev.close = psp_close;
	pspdev.read = psp_read;
	pspdev.write = psp_write;
	pspdev.lseek = psp_lseek;
	pspdev.stat = psp_stat;
	pspdev.dopen = psp_dopen;
	pspdev.dread = psp_dread;
	pspdev.dclose = psp_dclose;
	
	device_register(&pspdev, "/psp", NULL, -1);

	return 0;
}
