/*
 * String functions (string.c)
 * Part of libpogo, a c-library replacent for GBA
 * Programmed by Jonas Minnberg (Sasq)
 *
 * DESCRIPTION
 * Implements many of the standard c-library string functions.
 *
 **/

#include "stdcore.h"

int isdigit(char c)
{
	return (c >= '0' && c <= '9');
}

int isalpha(char c)
{
	return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z');
}

int isalnum(char c)
{
	return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9');
}

int toupper(int c)
{
	if(c >= 'a' && c <= 'z')
		return (c+('A'-'a'));
	else
		return c;
}

int tolower(int c)
{
	if(c >= 'A' && c <= 'Z')
		return (c+('a'-'A'));
	else
		return c;
}

int strtol(const char *s, char **endptr, int base)
{
	int val = 0;
	int v = 0;
	uchar neg = 0;

	while(*s == ' ')
		*s++;

	if(*s == '-') {
		s++;
		neg = 1;
	}

	if((*s == '0') && (s[1] == 'x'))
	{
		if(base == 0 || base == 16)
		{
			base = 16;
			s += 2;
		}
	}

	if(!base)
		base = 10;

	do
	{
		val = *s;
		if(val <= 0x39)
			val -= 0x30;
		else
			val = toupper(val) - 'A' + 10;

		if((val >= 0) && (val < base))
		{
			v *= base;
			v += val;
			s++;
		}
	}
	while((val >= 0) && (val < base));

	if(neg)
		v = -v;

	if(endptr)
		*endptr = (char *)s;

	return v;
}

int atoi(const char *s)
{
	 return strtol(s, (char **)NULL, 10);
}

char *itoa_size(int v, char *dst, unsigned int base, int size, char fill_char)
{
	unsigned int uv;
	static const char digits[] = "0123456789ABCDEF";
	char tmp[34];
	char *p = &tmp[33];
	uchar neg = 0;
	
	if(!size) size = 1;

	if(!v) {
		p = dst;
		while(size--)
			*p++ = '0';
		*p = 0;
		return dst;
	}

	*p-- = 0;
	if(v < 0 && base == 10) {
		uv = -v;
		neg = 1;
	} else
		uv = v;

	while(uv >= base) {
		*p-- = digits[uv % base];
		size--;
		uv /= base;
	}
	*p = digits[uv];
	size--;
	while(size > 0) {
		*(--p) = fill_char;
		size--;
	}

	if(neg)
		*(--p) = '-';
	strcpy(dst, p);

	return dst;
}

char *itoa(int v, char *dst, unsigned int base)
{
	return itoa_size(v, dst, base, 0, ' ');
}


char *strchr(const char *str, int c)
{
	while(*str && *str != c) str++;
	if(*str)
		return (char *)str;
	else
		return NULL;

}

char *strrchr(const char *str, int c)
{
	const char *p = NULL;
	while(*str)
	{
		if(*str == c)
			p = str;
		str++;
	}
	return (char *)p;
}

char *strcat(char *dst, const char *src)
{
	const char *s = src;
	char *d = dst;
	while(*d) d++;
	while(*s)
		*d++ = *s++;
	*d = 0;
	return dst;
}

int islower(int c)
{
	return (c >= 'a' && c <= 'z');
}

int isupper(int c)
{
	return (c >= 'A' && c <= 'Z');
}

int isprint(int c)
{
	return (c >= ' ' && c <= 127);
}

int stricmp(const char *s1, const char *s2)
{
	while(*s1 && (toupper(*s1) == toupper(*s2)))
	{
		s1++;
		s2++;
	}
	return (toupper(s1[0]) - toupper(s2[0]));
}

int strnicmp(const char *s1, const char *s2, int n)
{
	while(*s1 && --n && (toupper(*s1) == toupper(*s2)))
	{
		s1++;
		s2++;
	}

	return (toupper(s1[0]) - toupper(s2[0]));
}

char *strupr(char *str)
{
	char *p = str;
	while(*p)
	{
		*p = toupper(*p);
		p++;
	}

	return str;
}

char *strdup(const char *s)
{
	char *d = malloc(strlen(s)+1);
	strcpy(d, s);
	return d;
}
