#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pogo/types.h>
#include <pogo/font.h>

#include <ft2build.h>
#include FT_FREETYPE_H

//Settings 

#define BOLD 1
#define ITALIC 2
#define SOLID 4

char *outname = NULL;
char outdir[128] = "";
char *bmpname = NULL;
	
int write32(FILE *fp, int v)
{
	uchar t[4];
	t[0] = v & 0xff;
	t[1] = (v>>8) & 0xff;
	t[2] = (v>>16) & 0xff;
	t[3] = v>>24;
	return fwrite(t, 1, 4, fp);
}

int write16(FILE *fp, int v)
{
	uchar t[2];
	t[0] = v & 0xff;
	t[1] = v>>8;
	return fwrite(t, 1, 2, fp);
}

int write8(FILE *fp, int v)
{
	uchar t = v;
	return fwrite(&t, 1, 1, fp);
}

void blit(uchar *dest, int dpitch, uchar *src, int width, int height, int pitch)
{
	int dmod = dpitch - width;
	int smod = pitch - width;
	while(height--)
	{
		int w = width;
		while(w--)
		{
			*dest++ = *src++;
		}

		dest += dmod;
		src += smod;
	}
}

void blit_mono(uchar *dest, int dpitch, uchar *src, int width, int height, int pitch)
{
	int dmod = dpitch - width;
	int smod = pitch - width/8;
	while(height--)
	{
		int w = width;
		int b = 7;
		while(w--)
		{
			*dest++ = (*src & (1<<b)) ? 0xff : 0x00;
			b--;
			if(b == -1)
			{
				src++;
				b = 7;
			}
			//w -= 8;
		}

		dest += dmod;
		src += smod;
	}
}

char *makeBaseName(char *name, char *dest)
{
	char *start, *end, *p;
	if((p = strrchr(name, '/')))
		start = p+1;
	else
	if((p = strrchr(name, '\\')))
		start = p+1;
	else
		start = name;
		
	if((p = strrchr(name, '.')))
		end = p;
	else
		end = &name[strlen(name)];
	
	strncpy(dest, start, end-start);
	dest[end-start] = 0;
	return dest;
}

void ConvertFont(char *filename, int size, int flags, int *sequence)
{
	int i,x;
	FT_Library library;
	FT_Face face;
	FT_Error error;

	error = FT_Init_FreeType(&library);

	if(error)
		return;

	//char ttfname[32];
	//sprintf(ttfname, "%s.ttf", fontname);
/*
	i=0;
	while(!error)
	{
		error = FT_New_Face(library, fontname, i, &face);
		if(!error)
		{
			printf("Font is %s/%s\n", face->family_name, face->style_name);
			FT_Done_Face(face);
			i++;
		}
	}

	return;
*/
	error = FT_New_Face(library, filename, 0, &face);

	error = FT_Set_Pixel_Sizes(face, 0, size);
	printf("\n** Opening %s (%d)\n", filename, size);
	
	char fontname[80];
	strcpy(fontname, face->family_name);
	char *p;
	if(p = strchr(fontname, ' '))
		*p = 0;
	
	if(face->style_flags & FT_STYLE_FLAG_BOLD)
		flags |= BOLD;

	if(face->style_flags & FT_STYLE_FLAG_ITALIC)
		flags |= ITALIC;
	
	printf("Font is %s/%s (%d/%d high) / %d\n", face->family_name, face->style_name, face->ascender - face->descender, face->height, face->units_per_EM);
	//printf("max advance %d %d\n", face->max_advance_width, face->max_advance_height);

	int asc = (face->ascender * size + 1023) / face->units_per_EM;

	int height = ((face->ascender - face->descender) * size + 2047) / face->units_per_EM;
	
	
	int total = 0;
	int count = 0;
	int lowBearing = 0;
	int lowRight = 0;
	for(i=0; sequence[i] > 0; i++)
	{
		error = FT_Load_Char( face, sequence[i], FT_LOAD_DEFAULT );
		x = (face->glyph->advance.x+63)/64;
		total += x;
		count++;

		int right = face->glyph->advance.x - face->glyph->metrics.width - face->glyph->metrics.horiBearingX;

		if(right < lowRight)
			lowRight = right;

		if(face->glyph->metrics.horiBearingX < lowBearing)
			lowBearing = face->glyph->metrics.horiBearingX;

		//printf("%c : %d %d %d\n", sequence[i], face->glyph->metrics.horiBearingX, face->glyph->metrics.horiAdvance, face->glyph->metrics.width);

	}
	
	printf("lowBearing %d\n", lowBearing);

	uint16 *offsets = (uint16 *)malloc(2 * (count+2));

	lowBearing = (-lowBearing+63)/64;

	lowRight = (-lowRight+63)/64;

	//printf("Fontsize %d x %d + %d for every char\n", total, height, lowBearing);

	total += (count * (lowBearing+lowRight));
	total = ((total+3) & 0xFFFFFFFC);

	uchar *pixels = (uchar *)malloc(total*(height+20));
	pixels += (total*10);
	memset(pixels, 0, total*height);


	uchar *ptr = pixels + asc * total;

	// ptr points at baseline

	error = FT_Load_Char( face, ' ', FT_LOAD_DEFAULT );
	int space_size = ((face->glyph->advance.x+63)/64 + lowBearing + lowRight);


	int offs = 0;
	
	bool fixed = true;
	int cwidth = -1;

	for(i=0; sequence[i] > 0; i++)
	{
		//error = FT_Load_Char( face, sequence[i], FT_LOAD_RENDER );
		if(flags & SOLID)
		{
			error = FT_Load_Char( face, sequence[i], FT_LOAD_MONOCHROME);
			error = FT_Render_Glyph(face->glyph, FT_RENDER_MODE_MONO);
		}
		else
			error = FT_Load_Char( face, sequence[i], FT_LOAD_RENDER );

		x = (face->glyph->advance.x+63)/64;
		FT_Bitmap *bitmap = &face->glyph->bitmap;

		uchar *dest = ptr + face->glyph->bitmap_left + lowBearing - (face->glyph->bitmap_top * total);
		if((dest < pixels) || ((dest + bitmap->rows * total) > (pixels + total*(height+1))))
		{
			//printf("ERROR\n");
		}

		if(flags & SOLID)
			blit_mono(dest, total, bitmap->buffer, bitmap->width, bitmap->rows, bitmap->pitch);
		else
			blit(dest, total, bitmap->buffer, bitmap->width, bitmap->rows, bitmap->pitch);
		ptr += (x + lowBearing+lowRight);
		
		//printf("%d\n", x + lowBearing+lowRight);
		
		if((cwidth != -1) && (cwidth != (x + lowBearing+lowRight)))
			fixed = false;
		
		cwidth = (x + lowBearing+lowRight);
		
		
		offsets[i] = offs;
		offs += cwidth;
		
		

		//if(bitmap->

		//printf("%c : %d/%d\n", sequence[i], (x + lowBearing), face->glyph->bitmap_left);// * size / face->units_per_EM);
		//bitmap->buffer[x+w*y];
	}
	offsets[i] = offs;

	FT_Done_Face(face);

	Font pfont;

	pfont.width = total;
	pfont.charwidth = space_size;
	pfont.height = height;
	pfont.spacing = -lowBearing-lowRight;

	pfont.first = sequence[0];
	pfont.last = sequence[count-1];
	
	//printf("From %d->%d\n", pfont.first, pfont.last);

	pfont.flags = (fixed ? 0 : FFLG_PROPORTIONAL) | FFLG_COLOR | FFLG_TRANSP | FFLG_UNICODE;

	char out[128];
	
	//sprintf(outname, "%s%d%s.bmp", fontname, size);
	//if(bmpname)
	//	SDL_SaveBMP(fontsurf, bmpname);

	//char basename[64];
	//makeBaseName(fontname, basename);

	sprintf(out, "%s%s%d%s%s.font", outdir, fontname, size,
			flags & BOLD ? "b" : "",
			flags & ITALIC ? "i" : ""
			);
			
	printf("Writing %s\n\n", out);
			
	if(!outname)
		outname = out;
	
	FILE *fp = fopen(outname, "wb");
	
	//uint16 cc = 0;
	
	fwrite(&pfont, 1, 6, fp);
	write16(fp, pfont.width);
	write16(fp, 0);
	write16(fp, 0);

	//SDL_LockSurface(fontsurf);
	fwrite(pixels, total, height, fp);
	//SDL_UnlockSurface(fontsurf);
	
	//int newseq[48];
	//int *nsp = newseq;
	
	for(i=0; sequence[i] > 0;)
	{
		//*nsp++ = sequence[i];
		write32(fp, sequence[i]);
		while(sequence[i+1] - sequence[i] == 1) i++;
		//*nsp++ = sequence[i++];
		write32(fp, sequence[i++]);
	}
	
	write32(fp, -1);
	write32(fp, -1);
	//*nsp++ = -1;
	//*nsp++ = -1;


	int j;
	//fwrite(newseq, sizeof(int), nsp-newseq, fp);		
	//fwrite(offsets, 1, (i+1)*2, fp);
	for(j=0; j<i+1; j++)
	{
		write16(fp, offsets[j]);
	}
	
	fclose(fp);

}

int main(int argc, char **argv)
{

	int i,x;
	int flags = 0;
	int size = 12;
	int tryflags = 0;
	char *p,*name = "";

	if(argc < 3)
	{
		printf("Usage: %s [-d<outdir>] [-s] <ttfname> <size>\n", argv[0]);
		return 0;
	}
	
	//name = argv[1];
	//size = atoi(argv[2]);

	for(i=1; i<argc; i++)
	{
		if(argv[i][0] == '-')
		{
			switch(argv[i][1])
			{
			case 's':
				flags |= SOLID;
				break;
			case 'B':
				if(argv[i][2] && argv[i][2] != ' ')
					bmpname = &argv[i][2];
				else
					bmpname = "font.bmp";
				break;
			case 'b':
				tryflags |= BOLD;
				break;
			case 'i':
				tryflags |= ITALIC;
				break;
			//case 'y':
			//	tweaky = atoi(&argv[i][2]);
				break;
			case 'o':
				outname = &argv[i][2];
				break;
			case 'd':
				strcpy(outdir,&argv[i][2]);
				p = &outdir[strlen(outdir)-1];
				if((*p != '/') && (*p != '\\'))
				{
					p[1] = '/';
					p[2] = 0;
				}
				break;
			}
		}
		else
		if( (x = atoi(argv[i])) > 0 )
		{
			size = x;
		}
		else
			name = argv[i];
	}
	
	int sequence[1024];
	int *s = sequence;
	
	// ASCII
	for(i=0x20; i<=0x7E; i++)
		*s++ = i;

	//*s++ = 151;

	// Copyright
	*s++ = 0xA9;

	// Latin-1
	for(i=0xC0; i<=0xFF; i++)
		*s++ = i;

	// Unicode Long dashes
	for(i=0x2013; i<=0x2014; i++)
		*s++ = i;

	// Unicode Quotes
	for(i=0x2018; i<=0x201E; i++)
		*s++ = i;

	
	*s++ = 0x21e5;
	*s++ = 0x21b5;
	*s++ = 0x2190;
	*s++ = 0x21e7;
	
	// Bullet	
	*s++ = 0x2219;
			
	
	*s++ = -1;
	*s++ = -1;
	
	printf("%s %d\n", name, size);
	
	
	
	ConvertFont(name, size, flags, sequence);
	
	return 0;	
}
